# -*- coding: utf-8 -*-
#
# Copyright (C) 2019-2020 CERN.
# Copyright (C) 2019-2020 RERO.
#
# Invenio-Circulation is free software; you can redistribute it and/or modify
# it under the terms of the MIT License; see LICENSE file for more details.

"""Tests for loan update."""

import json

from flask import url_for

from invenio_circulation.api import get_loan_for_item


def _post(app, json_headers, loan, payload):
    with app.test_client() as client:
        url = url_for(
            "invenio_circulation_loan_replace_item.loan_replace_item_resource",
            pid_value=loan["pid"],
        )
        res = client.post(url, headers=json_headers, data=json.dumps(payload))
        data = json.loads(res.data.decode("utf-8"))
    return res, data


def test_loan_replace_item(app, json_headers, indexed_loans):
    """Test that no Loan is returned for the given Item if only pendings."""
    item_pid = dict(type="itemid", value="item_on_loan_2")
    loan = get_loan_for_item(item_pid)
    payload = dict(item_pid=dict(type="itemid", value="new_item_pid"))
    res, data = _post(app, json_headers, loan, payload)
    assert res.status == "202 ACCEPTED"
    assert data["metadata"]["item_pid"] == payload["item_pid"]
    # the ref won't change as it is pointing back to loan (it will
    # resolve by loan endpoint)
    assert data["metadata"]["item"]["ref"] == loan["pid"]


def test_loan_replace_item_inactive_state(app, json_headers, indexed_loans):
    """Test item replacement on a Loan that is not active."""
    for _pid, _loan in indexed_loans:
        if _loan["state"] == "ITEM_RETURNED":
            loan = _loan
            break
    payload = dict(item_pid=dict(type="itemid", value="new_item_pid"))
    res, data = _post(app, json_headers, loan, payload)
    assert data["status"] == 400
    assert data["message"] == (
        "Cannot replace item in a loan that is not in active state. "
        "Current loan state '{}'".format(loan["state"])
    )


def test_loan_replace_item_wo_params(app, json_headers, indexed_loans):
    """Test that no Loan is returned for the given Item if only pendings."""
    item_pid = dict(type="itemid", value="item_on_loan_2")
    loan = get_loan_for_item(item_pid)
    payload = {}
    res, data = _post(app, json_headers, loan, payload)
    assert res.status_code == 400


def test_loan_replace_item_wo_loan(app, json_headers, indexed_loans):
    """Test that no Loan is returned for the given Item if only pendings."""
    loan = {"pid": "not_existing_one"}
    payload = dict(type="itemid", value="item_on_loan_2")
    res, data = _post(app, json_headers, loan, payload)
    assert data["status"] == 404



# -*- coding: utf-8 -*-
#
# Copyright (C) 2018-2020 CERN.
# Copyright (C) 2018-2020 RERO.
#
# Invenio-Circulation is free software; you can redistribute it and/or modify
# it under the terms of the MIT License; see LICENSE file for more details.

"""Tests for loan states."""

from datetime import timedelta

import arrow
import pytest

from invenio_circulation.errors import ItemNotAvailableError, \
    TransitionConstraintsViolationError
from invenio_circulation.proxies import current_circulation

from .helpers import SwappedConfig


def test_created_to_item_on_loan_available_item_with_default_location(
    loan_created, params, mock_is_item_available_for_checkout
):
    """Test direct checkout on available item with default location."""

    mock_is_item_available_for_checkout.return_value = True

    assert loan_created["state"] == "CREATED"

    with SwappedConfig(
        "CIRCULATION_ITEM_LOCATION_RETRIEVER",
        lambda x: "pickup_location_pid"
    ):
        loan = current_circulation.circulation.trigger(
            loan_created, **dict(params, trigger="checkout")
        )

    assert loan["state"] == "ITEM_ON_LOAN"
    assert loan["pickup_location_pid"] == "pickup_location_pid"
    assert loan["item_pid"] == dict(type="itemid", value="item_pid")


def test_created_to_item_on_loan_available_item_with_specified_location(
    loan_created, params, mock_is_item_available_for_checkout
):
    """Test direct checkout on available item with different location."""

    mock_is_item_available_for_checkout.return_value = True

    assert loan_created["state"] == "CREATED"

    with SwappedConfig(
        "CIRCULATION_ITEM_LOCATION_RETRIEVER",
        lambda x: "pickup_location_pid"
    ):
        loan = current_circulation.circulation.trigger(
            loan_created, **dict(params, trigger="checkout",
                                 pickup_location_pid="other_location_pid")
        )

    assert loan["state"] == "ITEM_ON_LOAN"
    assert loan["pickup_location_pid"] == "other_location_pid"
    assert loan["item_pid"] == dict(type="itemid", value="item_pid")


def test_created_to_item_on_loan_unavailable_item(
    loan_created, params, mock_ensure_item_is_available_for_checkout
):
    """Test direct checkout on unavailable item."""

    assert loan_created["state"] == "CREATED"

    mock_ensure_item_is_available_for_checkout.side_effect = \
        ItemNotAvailableError(item_pid=dict(type="itemid", value="1"),
                              transition="Fake Transition")

    with pytest.raises(ItemNotAvailableError):
        with SwappedConfig(
            "CIRCULATION_ITEM_LOCATION_RETRIEVER",
            lambda x: "pickup_location_pid"
        ):
            current_circulation.circulation.trigger(
                loan_created, **dict(params, trigger="checkout",
                                     pickup_location_pid="other_location_pid")
            )


def test_created_to_item_on_loan_available_item_with_invalid_duration(
    loan_created, params, mock_is_item_available_for_checkout
):
    """Test direct checkout on available item with invalid duration."""

    mock_is_item_available_for_checkout.return_value = True

    assert loan_created["state"] == "CREATED"

    params["start_date"] = arrow.get("2018-01-01")
    params["end_date"] = params["start_date"] + timedelta(days=60)

    with pytest.raises(TransitionConstraintsViolationError):
        with SwappedConfig(
            "CIRCULATION_ITEM_LOCATION_RETRIEVER",
            lambda x: "pickup_location_pid"
        ):
            current_circulation.circulation.trigger(
                loan_created, **dict(params, trigger="checkout")
            )


def test_created_to_item_on_loan_available_item_with_valid_duration(
    loan_created, params, mock_ensure_item_is_available_for_checkout
):
    """Test direct checkout on available item with valid duration."""

    mock_ensure_item_is_available_for_checkout.side_effect = None

    assert loan_created["state"] == "CREATED"

    params["transaction_date"] = arrow.utcnow()
    params["start_date"] = arrow.get("2018-01-01")
    params["end_date"] = params["start_date"] + timedelta(days=59)

    with SwappedConfig(
        "CIRCULATION_ITEM_LOCATION_RETRIEVER",
        lambda x: "pickup_location_pid"
    ):
        loan = current_circulation.circulation.trigger(
            loan_created, **dict(params, trigger="checkout")
        )

    assert loan["state"] == "ITEM_ON_LOAN"
    assert loan["pickup_location_pid"] == "pickup_location_pid"
    assert loan["item_pid"] == dict(type="itemid", value="item_pid")
    assert loan["transaction_date"] == params["transaction_date"].isoformat()


def test_pending_to_item_on_loan_available_item(
    loan_created, params, mock_ensure_item_is_available_for_checkout
):
    """Test direct checkout on available item."""

    mock_ensure_item_is_available_for_checkout.side_effect = None

    with SwappedConfig(
        "CIRCULATION_ITEM_LOCATION_RETRIEVER",
        lambda x: "pickup_location_pid"
    ):
        loan = current_circulation.circulation.trigger(
            loan_created, **dict(params, trigger="request")
        )

        assert loan["state"] == "PENDING"

        loan = current_circulation.circulation.trigger(
            loan_created, **dict(params, trigger="checkout",
                                 pickup_location_pid="other_location_pid")
        )

        assert loan["state"] == "ITEM_ON_LOAN"
        assert loan["pickup_location_pid"] == "other_location_pid"
        assert loan["item_pid"] == dict(type="itemid", value="item_pid")



# -*- coding: utf-8 -*-
#
# Copyright (C) 2018-2019 CERN.
# Copyright (C) 2018-2019 RERO.
#
# Invenio-Circulation is free software; you can redistribute it and/or modify
# it under the terms of the MIT License; see LICENSE file for more details.

"""Tests for loan states."""

from datetime import timedelta

import arrow
import pytest

from invenio_circulation.errors import TransitionConstraintsViolationError
from invenio_circulation.proxies import current_circulation

from .helpers import SwappedConfig


def test_checkout_from_item_at_desk_valid_duration(loan_created, params):
    """Test transition from ITEM_AT_DESK to ITEM_ON_LOAN state."""
    loan = current_circulation.circulation.trigger(
        loan_created,
        **dict(
            params,
            trigger="request",
            pickup_location_pid="pickup_location_pid",
        )
    )
    assert loan["state"] == "PENDING"

    with SwappedConfig(
        "CIRCULATION_ITEM_LOCATION_RETRIEVER", lambda x: "pickup_location_pid"
    ):
        loan = current_circulation.circulation.trigger(
            loan, **dict(params)
        )
        assert loan["state"] == "ITEM_AT_DESK"

        loan = current_circulation.circulation.trigger(
            loan, **dict(params)
        )

        assert loan["state"] == "ITEM_ON_LOAN"


def test_checkout_from_item_at_desk_invalid_duration(loan_created, params):
    """Test transition from ITEM_AT_DESK to ITEM_ON_LOAN state."""
    loan = current_circulation.circulation.trigger(
        loan_created,
        **dict(
            params,
            trigger="request",
            pickup_location_pid="pickup_location_pid",
        )
    )
    assert loan["state"] == "PENDING"

    with SwappedConfig(
        "CIRCULATION_ITEM_LOCATION_RETRIEVER", lambda x: "pickup_location_pid"
    ):
        loan = current_circulation.circulation.trigger(
            loan, **dict(params)
        )
        assert loan["state"] == "ITEM_AT_DESK"

        now = arrow.utcnow()
        end_date = now + timedelta(days=60)
        loan["start_date"] = now.date().isoformat()
        loan["end_date"] = end_date.date().isoformat()

        with pytest.raises(TransitionConstraintsViolationError):
            current_circulation.circulation.trigger(
                loan, **dict(params)
            )



# -*- coding: utf-8 -*-
#
# Copyright (C) 2019-2020 CERN.
# Copyright (C) 2019-2020 RERO.
#
# Invenio-Circulation is free software; you can redistribute it and/or modify
# it under the terms of the MIT License; see LICENSE file for more details.

"""Circulation API."""

from datetime import timedelta


def patron_exists(patron_pid):
    """Return True if patron exists, False otherwise."""
    return True


def item_exists(item_pid):
    """Return True if item exists, False otherwise."""
    return True


def document_exists(document_pid):
    """Return True if document exists, False otherwise."""
    return True


def item_can_circulate(item_pid):
    """Return True if the item can circulate, False otherwise."""
    return True


def can_be_requested(loan):
    """Return True if the given record can be requested, False otherwise."""
    return True


# NOTE: Its on purpose `ref` and not `$ref` so it doesn't try to resolve
def item_ref_builder(loan_pid, loan):
    """Return the $ref for given loan_pid."""
    return {"ref": "{}".format(loan_pid)}


# NOTE: Its on purpose `ref` and not `$ref` so it doesn't try to resolve
def patron_ref_builder(loan_pid, loan):
    """Return the $ref for given loan_pid."""
    return {"ref": "{}".format(loan_pid)}


# NOTE: Its on purpose `ref` and not `$ref` so it doesn't try to resolve
def document_ref_builder(loan_pid, loan):
    """Return the $ref for given loan_pid."""
    return {"ref": "{}".format(loan_pid)}


def item_location_retriever(item_pid):
    """Retrieve the location pid of the passed item pid."""
    return ""


def get_default_loan_duration(loan, initial_loan):
    """Return a default loan duration in timedelta."""
    return timedelta(days=30)


def get_default_extension_duration(loan, initial_loan):
    """Return a default extension duration in timedelta."""
    return timedelta(days=30)


def get_default_extension_max_count(loan):
    """Return a default extensions max count."""
    return float("inf")


def is_loan_duration_valid(loan):
    """Validate the loan duration."""
    return loan["end_date"] > loan["start_date"] and loan["end_date"] - loan[
        "start_date"
    ] < timedelta(days=60)


def transaction_location_validator(transaction_location_pid):
    """Validate that the given transaction location PID is valid."""
    return transaction_location_pid == "loc_pid"


def transaction_user_validator(transaction_user_pid):
    """Validate that the given transaction user PID is valid."""
    return transaction_user_pid == "user_pid"


def validate_item_pickup_transaction_locations(loan, destination, **kwargs):
    """Validate the loan item, pickup and transaction locations."""
    pickup_location_pid = loan["pickup_location_pid"]
    item_location_pid = kwargs["item_location_pid"]
    if destination == "ITEM_AT_DESK":
        return pickup_location_pid == item_location_pid
    elif destination == "ITEM_IN_TRANSIT_FOR_PICKUP":
        return pickup_location_pid != item_location_pid



# -*- coding: utf-8 -*-
#
# Copyright (C) 2018 CERN.
# Copyright (C) 2018 RERO.
#
# Invenio-Circulation is free software; you can redistribute it and/or modify
# it under the terms of the MIT License; see LICENSE file for more details.

"""Helper proxy to the state object."""

from __future__ import absolute_import, print_function

from flask import current_app
from werkzeug.local import LocalProxy

current_circulation = LocalProxy(
    lambda: current_app.extensions['invenio-circulation']
)
"""Helper proxy to circulation state object."""



# -*- coding: utf-8 -*-
#
# Copyright (C) 2019 CERN.
#
# Invenio-Circulation is free software; you can redistribute it and/or modify
# it under the terms of the MIT License; see LICENSE file for more details.

"""Tests for circulation signals."""

from invenio_circulation.proxies import current_circulation
from invenio_circulation.signals import loan_state_changed


def test_signals_loan_request(loan_created, params):
    """Test signal for loan request action."""
    recorded = []

    def record_signals(_, initial_loan, loan, trigger):
        recorded.append((initial_loan, loan, trigger))

    loan_state_changed.connect(record_signals, weak=False)

    assert len(recorded) == 0
    current_circulation.circulation.trigger(
        loan_created,
        **dict(
            params,
            trigger="request",
            pickup_location_pid="pickup_location_pid",
        )
    )
    assert len(recorded) == 1
    initial_loan, updated_loan, trigger = recorded.pop()
    assert initial_loan["state"] == "CREATED"
    assert updated_loan["state"] == "PENDING"
    assert trigger == "request"


def test_signals_loan_extend(loan_created, params):
    """Test signals for loan extend action."""
    recorded = []

    def record_signals(_, initial_loan, loan, trigger):
        recorded.append((initial_loan, loan, trigger))

    loan_state_changed.connect(record_signals, weak=False)

    assert len(recorded) == 0
    loan = current_circulation.circulation.trigger(
        loan_created, **dict(params, trigger="checkout")
    )
    assert len(recorded) == 1
    initial_loan, updated_loan, trigger = recorded.pop()
    assert initial_loan["state"] == "CREATED"
    assert updated_loan["state"] == "ITEM_ON_LOAN"
    assert trigger == "checkout"

    current_circulation.circulation.trigger(
        loan, **dict(params, trigger="extend")
    )
    assert len(recorded) == 1
    initial_loan, updated_loan, trigger = recorded.pop()
    assert initial_loan["state"] == "ITEM_ON_LOAN"
    assert updated_loan["state"] == "ITEM_ON_LOAN"
    assert initial_loan["end_date"] != updated_loan["end_date"]
    assert trigger == "extend"



# -*- coding: utf-8 -*-
#
# Copyright (C) 2018 CERN.
# Copyright (C) 2018 RERO.
#
# Invenio-Circulation is free software; you can redistribute it and/or modify
# it under the terms of the MIT License; see LICENSE file for more details.

"""Invenio module for the circulation of bibliographic items."""

from __future__ import absolute_import, print_function

from .ext import InvenioCirculation
from .version import __version__

__all__ = ('__version__', 'InvenioCirculation')



# -*- coding: utf-8 -*-
#
# Copyright (C) 2018-2020 CERN.
# Copyright (C) 2018-2020 RERO.
#
# Invenio-Circulation is free software; you can redistribute it and/or modify
# it under the terms of the MIT License; see LICENSE file for more details.

"""Tests for loan search class."""

from elasticsearch import VERSION as ES_VERSION

from invenio_circulation.api import Loan
from invenio_circulation.search.api import search_by_patron_item_or_document, \
    search_by_patron_pid, search_by_pid


def _assert_total(total, expected):
    """Assert total (ES6 compatibility)."""
    if ES_VERSION[0] >= 7:
        assert total.value == expected
    else:
        assert total == expected


def test_search_loans_by_pid(indexed_loans):
    """Test retrieve loan list belonging to an item."""
    item_pid = dict(type="itemid", value="item_pending_1")
    loans = list(
        search_by_pid(
            item_pid=item_pid
        ).scan()
    )
    assert len(loans) == 1
    loan = Loan.get_record_by_pid(loans[0]["pid"])
    assert loan["item_pid"] == item_pid


def test_search_loans_by_pid_filtering_states(indexed_loans):
    """Test retrieve loan list belonging to an item filtering states."""
    search = search_by_pid(
        item_pid=dict(type="itemid", value="item_multiple_pending_on_loan_7"),
        filter_states=["PENDING", "ITEM_ON_LOAN"],
    )
    search_result = search.execute()
    _assert_total(search_result.hits.total, 3)


def test_search_loans_by_pid_excluding_states(indexed_loans):
    """Test retrieve loan list belonging to an item excluding states."""
    search_result = search_by_pid(
        item_pid=dict(type="itemid", value="item_multiple_pending_on_loan_7"),
        exclude_states=["ITEM_ON_LOAN"],
    ).execute()
    _assert_total(search_result.hits.total, 2)


def test_search_loans_by_patron_pid(indexed_loans):
    """Test retrieve loan list belonging to a patron."""
    search_result = search_by_patron_pid("1").execute()
    _assert_total(search_result.hits.total, 8)

    search_result = search_by_patron_pid("2").execute()
    _assert_total(search_result.hits.total, 3)

    search_result = search_by_patron_pid("3").execute()
    _assert_total(search_result.hits.total, 1)


def test_search_loans_by_patron_and_item_or_document(indexed_loans):
    """Test retrieve loan list by patron and items."""
    search_result = search_by_patron_item_or_document(
        patron_pid="1", item_pid=dict(type="itemid", value="item_returned_3")
    ).execute()
    _assert_total(search_result.hits.total, 1)

    search_result = search_by_patron_item_or_document(
        patron_pid="1", item_pid=dict(type="itemid", value="not_existing")
    ).execute()
    _assert_total(search_result.hits.total, 0)

    search_result = search_by_patron_item_or_document(
        patron_pid="999999",
        item_pid=dict(type="itemid", value="item_returned_3"),
    ).execute()
    _assert_total(search_result.hits.total, 0)

    search_result = search_by_patron_item_or_document(
        patron_pid="1", document_pid="document_pid"
    ).execute()
    _assert_total(search_result.hits.total, 8)

    search_result = search_by_patron_item_or_document(
        patron_pid="1", document_pid="not_existing"
    ).execute()
    _assert_total(search_result.hits.total, 0)

    search_result = search_by_patron_item_or_document(
        patron_pid="999999", document_pid="document_returned_1"
    ).execute()
    _assert_total(search_result.hits.total, 0)


def test_search_loans_by_patron_and_item_or_document_filtering_states(
    indexed_loans
):
    """Test retrieve loan list by patron and items filtering states."""
    search = search_by_patron_item_or_document(
        patron_pid="1",
        item_pid=dict(type="itemid", value="item_returned_3"),
        filter_states=["ITEM_RETURNED"],
    )
    search_result = search.execute()
    _assert_total(search_result.hits.total, 1)

    search = search_by_patron_item_or_document(
        patron_pid="1",
        item_pid=dict(type="itemid", value="item_returned_3"),
        filter_states=["ITEM_AT_DESK"],
    )
    search_result = search.execute()
    _assert_total(search_result.hits.total, 0)

    search = search_by_patron_item_or_document(
        patron_pid="1",
        document_pid="document_pid",
        filter_states=["ITEM_RETURNED"],
    )
    search_result = search.execute()
    _assert_total(search_result.hits.total, 2)

    search = search_by_patron_item_or_document(
        patron_pid="2",
        document_pid="document_pid",
        filter_states=["ITEM_RETURNED"],
    )
    search_result = search.execute()
    _assert_total(search_result.hits.total, 1)

    search = search_by_patron_item_or_document(
        patron_pid="1",
        document_pid="document_pid",
        filter_states=["ITEM_AT_DESK"],
    )
    search_result = search.execute()
    _assert_total(search_result.hits.total, 1)

    search = search_by_patron_item_or_document(
        patron_pid="1", document_pid="document_pid", filter_states=["PENDING"]
    )
    search_result = search.execute()
    _assert_total(search_result.hits.total, 3)



# -*- coding: utf-8 -*-
#
# Copyright (C) 2019-2020 CERN.
# Copyright (C) 2019-2020 RERO.
#
# invenio-circulation is free software; you can redistribute it and/or modify
# it under the terms of the MIT License; see LICENSE file for more details.

"""Circulation record loaders module."""

from invenio_records_rest.loaders import marshmallow_loader

from .schemas.json import LoanReplaceItemSchemaV1, LoanSchemaV1

loan_loader = marshmallow_loader(LoanSchemaV1)
loan_replace_item_loader = marshmallow_loader(LoanReplaceItemSchemaV1)



# -*- coding: utf-8 -*-
#
# Copyright (C) 2018 CERN.
# Copyright (C) 2018 RERO.
#
# Invenio-Circulation is free software; you can redistribute it and/or modify
# it under the terms of the MIT License; see LICENSE file for more details.

"""Tests for loan states."""

from invenio_circulation.proxies import current_circulation

from .helpers import SwappedConfig


def test_item_on_loan_to_item_in_transit_to_house(
    loan_created, db, params
):
    """Test transition from ITEM_ON_LOAN to ITEM_IN_TRANSIT_TO_HOUSE state."""
    loan = current_circulation.circulation.trigger(
        loan_created,
        **dict(
            params,
            trigger="checkout",
            pickup_location_pid="loc_pid",
        )
    )

    assert loan["state"] == "ITEM_ON_LOAN"
    assert loan["transaction_location_pid"] == "loc_pid"

    # item is not returned to the same location
    with SwappedConfig(
        "CIRCULATION_ITEM_LOCATION_RETRIEVER", lambda x: "other_location_pid"
    ):
        loan = current_circulation.circulation.trigger(
            loan, **dict(params)
        )

        assert loan["state"] == "ITEM_IN_TRANSIT_TO_HOUSE"


def test_item_on_loan_to_item_returned_same_location(
    loan_created, db, params
):
    """Test transition from ITEM_ON_LOAN to ITEM_RETURNED state."""
    loan = current_circulation.circulation.trigger(
        loan_created,
        **dict(
            params,
            trigger="checkout",
            pickup_location_pid="loc_pid",
        )
    )

    assert loan["state"] == "ITEM_ON_LOAN"
    assert loan["transaction_location_pid"] == "loc_pid"

    # item is returned to the same location
    with SwappedConfig(
        "CIRCULATION_ITEM_LOCATION_RETRIEVER", lambda x: "loc_pid"
    ):
        loan = current_circulation.circulation.trigger(
            loan, **dict(params)
        )
        assert loan["state"] == "ITEM_RETURNED"



# -*- coding: utf-8 -*-
#
# Copyright (C) 2018-2019 CERN.
# Copyright (C) 2018-2019 RERO.
#
# Invenio-Circulation is free software; you can redistribute it and/or modify
# it under the terms of the MIT License; see LICENSE file for more details.

"""Module tests."""

from __future__ import absolute_import, print_function


def test_version():
    """Test version import."""
    from invenio_circulation import __version__

    assert __version__



# -*- coding: utf-8 -*-
#
# Copyright (C) 2018-2020 CERN.
# Copyright (C) 2018-2020 RERO.
#
# Invenio-Circulation is free software; you can redistribute it and/or modify
# it under the terms of the MIT License; see LICENSE file for more details.

"""Invenio Circulation custom transitions."""

from flask import current_app
from invenio_db import db

from invenio_circulation.proxies import current_circulation

from ..api import can_be_requested, get_available_item_by_doc_pid, \
    get_document_pid_by_item_pid, get_pending_loans_by_doc_pid, \
    is_item_at_desk_available_for_checkout
from ..errors import ItemDoNotMatchError, ItemNotAvailableError, \
    LoanMaxExtensionError, RecordCannotBeRequestedError, \
    TransitionConditionsFailedError, TransitionConstraintsViolationError
from ..transitions.base import Transition
from ..transitions.conditions import is_same_location


def _ensure_valid_loan_duration(loan, initial_loan):
    """Validate start and end dates for a loan."""
    loan.setdefault("start_date", loan["transaction_date"])

    if not loan.get("end_date"):
        get_loan_duration = current_app.config["CIRCULATION_POLICIES"][
            "checkout"
        ]["duration_default"]
        duration = get_loan_duration(loan, initial_loan)
        loan["end_date"] = loan["start_date"] + duration

    is_duration_valid = current_app.config["CIRCULATION_POLICIES"]["checkout"][
        "duration_validate"
    ]
    if not is_duration_valid(loan):
        msg = "The loan duration from '{0}' to '{1}' is not valid.".format(
            loan["start_date"].isoformat(), loan["end_date"].isoformat()
        )
        raise TransitionConstraintsViolationError(description=msg)


def _ensure_item_attached_to_loan(loan):
    """Validate that an item is attached to a loan."""
    if not loan.get("item_pid"):
        msg = "No item assigned to loan '{0}'.".format(loan.id)
        raise TransitionConditionsFailedError(description=msg)


def ensure_same_item(f):
    """Validate that the item PID exists and cannot be changed."""
    def inner(self, loan, initial_loan, **kwargs):
        item_pid = kwargs.get("item_pid")

        if item_pid:
            if not current_app.config["CIRCULATION_ITEM_EXISTS"](item_pid):
                msg = "Item '{0}:{1}' not found in the system".format(
                    item_pid["type"], item_pid["value"]
                )
                raise ItemNotAvailableError(description=msg)

            wrong_pid_value = loan.get("item_pid") and \
                item_pid["value"] != loan["item_pid"]["value"]
            wrong_pid_type = loan.get("item_pid") and \
                item_pid["type"] != loan["item_pid"]["type"]
            if wrong_pid_value or wrong_pid_type:
                msg = (
                    "Cannot change item '{0}:{1}' while performing an "
                    "action on this loan".format(
                        item_pid["type"], item_pid["value"]
                    )
                )
                raise ItemDoNotMatchError(description=msg)

        return f(self, loan, initial_loan, **kwargs)

    return inner


def _update_document_pending_request_for_item(item_pid, **kwargs):
    """Update pending loans on a Document with no Item attached yet.

    :param item_pid: a dict containing `value` and `type` fields to
        uniquely identify the item.
    """
    document_pid = get_document_pid_by_item_pid(item_pid)
    for pending_loan in get_pending_loans_by_doc_pid(document_pid):
        pending_loan["item_pid"] = item_pid
        pending_loan.commit()
        db.session.commit()
        current_circulation.loan_indexer().index(pending_loan)


def _ensure_same_location(item_pid, location_pid, destination, error_msg):
    """Validate that item location is same as given location."""
    if not is_same_location(item_pid, location_pid):
        error_msg += " Transition to '{}' has failed.".format(destination)
        raise TransitionConditionsFailedError(description=error_msg)


def _ensure_not_same_location(item_pid, location_pid, destination, error_msg):
    """Validate that item location is not the same as given location."""
    if is_same_location(item_pid, location_pid):
        error_msg += " Transition to '{}' has failed.".format(destination)
        raise TransitionConditionsFailedError(description=error_msg)


def _validate_item_pickup_transaction_locations(loan, destination, **kwargs):
    """Validate the loan item, pickup and transaction locations."""
    item_location_pid = \
        current_app.config["CIRCULATION_ITEM_LOCATION_RETRIEVER"](
            loan["item_pid"])
    kwargs["item_location_pid"] = item_location_pid
    validate_item_pickup_transaction_locations = current_app.config[
        "CIRCULATION_LOAN_LOCATIONS_VALIDATION"]
    if not validate_item_pickup_transaction_locations(
            loan, destination, **kwargs):
        raise TransitionConditionsFailedError()


def _get_item_location(item_pid):
    """Retrieve Item location based on PID."""
    return current_app.config["CIRCULATION_ITEM_LOCATION_RETRIEVER"](item_pid)


def _ensure_default_pickup_location(loan, context):
    """Set default pickup location if no one."""
    if not context.get("pickup_location_pid") \
            or "pickup_location_pid" not in loan:
        loan['pickup_location_pid'] = _get_item_location(loan['item_pid'])


class ToItemOnLoan(Transition):
    """Action to checkout."""

    def before(self, loan, initial_loan, **kwargs):
        """Validate checkout action."""
        super().before(loan, initial_loan, **kwargs)

        self.ensure_item_is_available_for_checkout(loan)

        _ensure_default_pickup_location(loan, kwargs)

        _ensure_valid_loan_duration(loan, initial_loan)


class ItemAtDeskToItemOnLoan(Transition):
    """Check-out action to perform a loan when item ready at desk."""

    def before(self, loan, initial_loan, **kwargs):
        """Validate checkout action."""
        super().before(loan, initial_loan, **kwargs)

        self.ensure_at_desk_item_is_available_for_checkout(loan)

        _ensure_default_pickup_location(loan, kwargs)

        _ensure_valid_loan_duration(loan, initial_loan)

    def ensure_at_desk_item_is_available_for_checkout(self, loan):
        """Validate that an item at desk is available for checkout."""
        self._check_item_before_availability(loan)

        # patron_pid is mandatory for next steps
        if 'patron_pid' not in loan:
            msg = "Patron not set for loan with pid '{}'".format(loan['pid'])
            raise TransitionConstraintsViolationError(description=msg)

        is_available = is_item_at_desk_available_for_checkout(
            loan['item_pid'],
            loan['patron_pid']
        )
        if not is_available:
            raise ItemNotAvailableError(
                item_pid=loan['item_pid'], transition=self.dest)


def check_request_on_document(f):
    """Decorator to check if the request is on document."""
    def inner(self, loan, initial_loan, **kwargs):
        document_pid = kwargs.get("document_pid")
        if document_pid and not kwargs.get("item_pid"):
            if not can_be_requested(loan):
                msg = "Cannot create a request for the document '{}'".format(
                    loan.get("document_pid")
                )
                raise RecordCannotBeRequestedError(description=msg)

            if self.assign_item:
                available_item_pid = get_available_item_by_doc_pid(
                    document_pid
                )
                if available_item_pid:
                    kwargs["item_pid"] = available_item_pid

        if kwargs.get("item_pid") and not kwargs.get("pickup_location_pid"):
            # if no pickup location was specified in the request,
            # assign a default one
            kwargs["pickup_location_pid"] = _get_item_location(
                kwargs["item_pid"]
            )

        return f(self, loan, initial_loan, **kwargs)
    return inner


class CreatedToPending(Transition):
    """Action to request to loan an item."""

    def __init__(
        self, src, dest, trigger="next", permission_factory=None, **kwargs
    ):
        """Constructor."""
        super().__init__(
            src,
            dest,
            trigger=trigger,
            permission_factory=permission_factory,
            **kwargs
        )
        self.assign_item = kwargs.get("assign_item", True)

    @check_request_on_document
    def before(self, loan, initial_loan, **kwargs):
        """Check if the loan request can be created."""
        super().before(loan, initial_loan, **kwargs)

        if not can_be_requested(loan):
            msg = "Cannot create a request for the loan '{}'".format(loan)
            raise RecordCannotBeRequestedError(description=msg)


class PendingToItemAtDesk(Transition):
    """Validate pending request to prepare the item at desk of its location."""

    def before(self, loan, initial_loan, **kwargs):
        """Validate if the item is for this location or should transit."""
        super().before(loan, initial_loan, **kwargs)

        # check if a request on document has no item attached
        _ensure_item_attached_to_loan(loan)
        # validate the item, pickup and transaction locations of the loan
        _validate_item_pickup_transaction_locations(loan, self.dest, **kwargs)


class PendingToItemInTransitPickup(Transition):
    """Validate pending request to send the item to the pickup location."""

    def before(self, loan, initial_loan, **kwargs):
        """Validate if the item is for this location or should transit."""
        super().before(loan, initial_loan, **kwargs)

        # check if a request on document has no item attached
        _ensure_item_attached_to_loan(loan)
        # validate the item, pickup and transaction locations of the loan
        _validate_item_pickup_transaction_locations(loan, self.dest, **kwargs)


class ItemOnLoanToItemOnLoan(Transition):
    """Extend action to perform a item loan extension."""

    def update_extension_count(self, loan):
        """Check number of extensions and update it."""
        extension_count = loan.get("extension_count", 0)
        extension_count += 1

        get_extension_max_count_func = current_app.config[
            "CIRCULATION_POLICIES"
        ]["extension"]["max_count"]
        extension_max_count = get_extension_max_count_func(loan)
        if extension_count > extension_max_count:
            raise LoanMaxExtensionError(
                loan_pid=loan["pid"], extension_count=extension_max_count
            )
        loan["extension_count"] = extension_count

    def update_loan_end_date(self, loan, initial_loan):
        """Update the end date of the extended loan."""
        get_extension_duration_func = current_app.config[
            "CIRCULATION_POLICIES"
        ]["extension"]["duration_default"]
        duration = get_extension_duration_func(loan, initial_loan)

        should_extend_from_end_date = current_app.config[
            "CIRCULATION_POLICIES"
        ]["extension"]["from_end_date"]
        if not should_extend_from_end_date:
            # extend from the transaction_date instead
            loan["end_date"] = loan["transaction_date"]

        loan["end_date"] += duration

    @ensure_same_item
    def before(self, loan, initial_loan, **kwargs):
        """Validate extension action."""
        super().before(loan, initial_loan, **kwargs)
        self.update_extension_count(loan)
        self.update_loan_end_date(loan, initial_loan)


class ItemOnLoanToItemInTransitHouse(Transition):
    """Check-in action when returning an item not to its belonging location."""

    @ensure_same_item
    def before(self, loan, initial_loan, **kwargs):
        """Validate check-in action."""
        super().before(loan, initial_loan, **kwargs)

        _ensure_not_same_location(
            loan["item_pid"],
            loan["transaction_location_pid"],
            self.dest,
            error_msg="Item should be returned (already in house).",
        )


class ItemOnLoanToItemReturned(Transition):
    """Check-in action when returning an item to its belonging location."""

    def __init__(
        self, src, dest, trigger="next", permission_factory=None, **kwargs
    ):
        """Constructor."""
        super().__init__(
            src,
            dest,
            trigger=trigger,
            permission_factory=permission_factory,
            **kwargs
        )
        self.assign_item = kwargs.get("assign_item", True)

    @ensure_same_item
    def before(self, loan, initial_loan, **kwargs):
        """Validate check-in action."""
        super().before(loan, initial_loan, **kwargs)
        _ensure_same_location(
            loan["item_pid"],
            loan["transaction_location_pid"],
            self.dest,
            error_msg="Item should be in transit to house.",
        )

        # set end loan date as transaction date when completing loan
        loan["end_date"] = loan["transaction_date"]

    def after(self, loan, initial_loan):
        """Check for pending requests on this item after check-in."""
        super().after(loan, initial_loan)
        if self.assign_item:
            _update_document_pending_request_for_item(loan["item_pid"])


class ItemInTransitHouseToItemReturned(Transition):
    """Check-in action when returning an item to its belonging location."""

    def __init__(
        self, src, dest, trigger="next", permission_factory=None, **kwargs
    ):
        """Constructor."""
        super().__init__(
            src,
            dest,
            trigger=trigger,
            permission_factory=permission_factory,
            **kwargs
        )
        self.assign_item = kwargs.get("assign_item", True)

    @ensure_same_item
    def before(self, loan, initial_loan, **kwargs):
        """Validate check-in action."""
        super().before(loan, initial_loan, **kwargs)

        _ensure_same_location(
            loan["item_pid"],
            loan["transaction_location_pid"],
            self.dest,
            error_msg="Item should be in transit to house.",
        )

    def after(self, loan, initial_loan):
        """Check for pending requests on this item after check-in."""
        super().after(loan, initial_loan)
        if self.assign_item:
            _update_document_pending_request_for_item(loan["item_pid"])


class ToCancelled(Transition):
    """When cancelling a loan, ensure that the item is not changed."""

    @ensure_same_item
    def before(self, loan, initial_loan, **kwargs):
        """Validate cancel action."""
        super().before(loan, initial_loan, **kwargs)



# -*- coding: utf-8 -*-
#
# Copyright (C) 2019-2020 CERN.
# Copyright (C) 2019-2020 RERO.
#
# invenio-circulation is free software; you can redistribute it and/or modify
# it under the terms of the MIT License; see LICENSE file for more details.

"""Circulation record loaders JSON schemas."""

from datetime import date, datetime

import arrow
from flask import current_app
from flask_babelex import lazy_gettext as _
from invenio_records_rest.schemas import RecordMetadataSchemaJSONV1
from invenio_records_rest.schemas.fields import PersistentIdentifier
from marshmallow import Schema, ValidationError, fields, post_load, validates


class DateTimeString(fields.DateTime):
    """Custom DateTime field to return a string representation."""

    def __init__(self, **kwargs):
        """Constructor."""
        kwargs.setdefault("validate", self.validate_timezone)
        super().__init__(**kwargs)

    def validate_timezone(self, value):
        """Validate that the passed timezone, if any, is UTC."""
        # strong validation of input datetime to require UTC timezone
        if (
            arrow.get(value).isoformat() !=
            arrow.get(value).to("utc").isoformat()
        ):
            raise ValidationError(_("Not a valid ISO-8601 UTC datetime."))

    def deserialize(self, value, attr=None, data=None, **kwargs):
        """Validate ISO8601 datetime input but return the string value."""
        # value and _value can be <marshmallow.missing>
        _value = super().deserialize(value, attr, data, **kwargs)
        # return the value as string after marshmallow validation
        # because Invenio does not support Python datetime JSON serializer yet
        if _value and type(_value) == datetime:
            return _value.isoformat()
        return _value


class DateString(fields.Date):
    """Custom Date field to return a string representation."""

    def deserialize(self, value, attr=None, data=None, **kwargs):
        """Validate ISO8601 date input but return the string value."""
        # value and _value can be <marshmallow.missing>
        _value = super().deserialize(value, attr, data, **kwargs)
        # return the value as string after marshmallow validation
        # because Invenio does not support Python datetime JSON serializer yet
        if _value and type(_value) == date:
            return _value.isoformat()
        return _value


class LoanItemPIDSchemaV1(Schema):
    """Loan item PID Schema."""

    type = fields.Str(required=True)
    value = fields.Str(required=True)


class LoanSchemaV1(RecordMetadataSchemaJSONV1):
    """Loan schema."""

    class Meta:
        """Meta attributes for the schema."""

        from marshmallow import EXCLUDE

        unknown = EXCLUDE

    def get_pid_field(self):
        """Return loan PID field name."""
        return "pid"

    cancel_reason = fields.Str()
    document_pid = fields.Str()
    end_date = DateString()
    extension_count = fields.Integer()
    item_pid = fields.Nested(LoanItemPIDSchemaV1)
    patron_pid = fields.Str(required=True)
    pid = PersistentIdentifier()
    pickup_location_pid = fields.Str()
    request_expire_date = DateString()
    request_start_date = DateString()
    start_date = DateString()
    transaction_date = DateTimeString()
    transaction_location_pid = fields.Str(required=True)
    transaction_user_pid = fields.Str(required=True)

    @validates("transaction_location_pid")
    def validate_transaction_location_pid(self, value,  **kwargs):
        """Validate transaction_location_pid field."""
        transaction_location_is_valid = current_app.config[
            "CIRCULATION_TRANSACTION_LOCATION_VALIDATOR"
        ]
        if not transaction_location_is_valid(value):
            raise ValidationError(
                _("The loan `transaction_location_pid` is not valid."),
                field_names=["transaction_location_pid"],
            )

    @validates("transaction_user_pid")
    def validate_transaction_user_pid(self, value, **kwargs):
        """Validate transaction_user_pid field."""
        transaction_user_is_valid = current_app.config[
            "CIRCULATION_TRANSACTION_USER_VALIDATOR"
        ]
        if not transaction_user_is_valid(value):
            raise ValidationError(
                _("The loan `transaction_user_pid` is not valid."),
                field_names=["transaction_user_pid"],
            )

    @post_load
    def postload_fields(self, data, **kwargs):
        """Post-load record fields update."""
        record = self.context.get("record")
        if record:
            record.update(data)
            return record
        else:
            return data


class LoanReplaceItemSchemaV1(RecordMetadataSchemaJSONV1):
    """Loan replace item schema."""

    class Meta:
        """Meta attributes for the schema."""

        from marshmallow import EXCLUDE

        unknown = EXCLUDE

    item_pid = fields.Nested(LoanItemPIDSchemaV1, required=True)



# -*- coding: utf-8 -*-
#
# Copyright (C) 2018-2020 CERN.
# Copyright (C) 2018-2020 RERO.
#
# Invenio-Circulation is free software; you can redistribute it and/or modify
# it under the terms of the MIT License; see LICENSE file for more details.

"""Circulation API."""


import arrow

from .errors import NotImplementedConfigurationError


def patron_exists(patron_pid):
    """Return True if patron exists, False otherwise."""
    raise NotImplementedConfigurationError(
        config_variable="CIRCULATION_PATRON_EXISTS"
    )


def item_exists(item_pid):
    """Return True if item exists, False otherwise.

    :param item_pid: a dict containing `value` and `type` fields to
        uniquely identify the item.
    """
    raise NotImplementedConfigurationError(
        config_variable="CIRCULATION_ITEM_EXISTS"
    )


def document_exists(document_pid):
    """Return True if document exists, False otherwise."""
    raise NotImplementedConfigurationError(
        config_variable="CIRCULATION_DOCUMENT_EXISTS"
    )


# NOTE: Its on purpose `ref` and not `$ref` so it doesn't try to resolve
def item_ref_builder(loan_pid, loan):
    """Return the $ref for item_pid."""
    raise NotImplementedConfigurationError(
        config_variable="CIRCULATION_ITEM_REF_BUILDER"
    )


# NOTE: Its on purpose `ref` and not `$ref` so it doesn't try to resolve
def patron_ref_builder(patron_pid, loan):
    """Return the $ref for patron_pid."""
    raise NotImplementedConfigurationError(
        config_variable="CIRCULATION_PATRON_REF_BUILDER"
    )


# NOTE: Its on purpose `ref` and not `$ref` so it doesn't try to resolve
def document_ref_builder(document_pid, loan):
    """Return the $ref for document_pid."""
    raise NotImplementedConfigurationError(
        config_variable="CIRCULATION_DOCUMENT_REF_BUILDER"
    )


def item_location_retriever(item_pid):
    """Retrieve the location pid of the passed item pid.

    :param item_pid: a dict containing `value` and `type` fields to
        uniquely identify the item.
    """
    raise NotImplementedConfigurationError(
        config_variable="CIRCULATION_ITEM_LOCATION_RETRIEVER"
    )


def item_can_circulate(item_pid):
    """Return if item is available for checkout.

    :param item_pid: a dict containing `value` and `type` fields to
        uniquely identify the item.
    """
    raise NotImplementedConfigurationError(
        config_variable="CIRCULATION_POLICIES.checkout.item_can_circulate"
    )


def can_be_requested(loan):
    """Should return True if document or item can be requested."""
    raise NotImplementedConfigurationError(
        config_variable="CIRCULATION_POLICIES.request.can_be_requested"
    )


def get_default_loan_duration(loan, initial_loan):
    """Return a default loan duration in timedelta.

    :param loan: the current loan to extend, updated with transition params
    :param initial_loan: the loan before applying transition params
    """
    raise NotImplementedConfigurationError(
        config_variable="CIRCULATION_POLICIES.checkout.duration_default"
    )


def is_loan_duration_valid(loan):
    """Validate the loan duration."""
    raise NotImplementedConfigurationError(
        config_variable="CIRCULATION_POLICIES.checkout.duration_validate"
    )


def get_default_extension_duration(loan, initial_loan):
    """Return a default extension duration in timedelta.

    :param loan: the current loan to extend, updated with transition params
    :param initial_loan: the loan before applying transition params
    """
    raise NotImplementedConfigurationError(
        config_variable="CIRCULATION_POLICIES.extension.duration_default"
    )


def get_default_extension_max_count(loan):
    """Return a default extensions max count."""
    raise NotImplementedConfigurationError(
        config_variable="CIRCULATION_POLICIES.extension.max_count"
    )


def transaction_location_validator(transaction_location_pid):
    """Validate that the given transaction location PID is valid.

    :param transaction_location_pid: the transaction location PID sent with
        the REST request.
    """
    raise NotImplementedConfigurationError(
        config_variable="CIRCULATION_TRANSACTION_LOCATION_VALIDATOR"
    )


def transaction_user_validator(transaction_user_pid):
    """Validate that the given transaction user PID is valid.

    :param transaction_user_pid: the transaction user PID sent with
        the REST request.
    """
    raise NotImplementedConfigurationError(
        config_variable="CIRCULATION_TRANSACTION_USER_VALIDATOR"
    )


def str2datetime(str_date):
    """Parse string date with timezone and return a datetime object."""
    return arrow.get(str_date).to('utc')


def validate_item_pickup_transaction_locations(loan, destination, **kwargs):
    """Validate the loan item, pickup and transaction locations.

    This also allow for extra validation if needed at the library level.

    :param loan: the loan
    :param destination: the destination of the loan

    :return: False if validation is not possible, otherwise True
    """
    raise NotImplementedConfigurationError(
        config_variable="CIRCULATION_LOAN_LOCATIONS_VALIDATION"
    )



# -*- coding: utf-8 -*-
#
# Copyright (C) 2018-2020 CERN.
# Copyright (C) 2018-2020 RERO.
#
# Invenio-Circulation is free software; you can redistribute it and/or modify
# it under the terms of the MIT License; see LICENSE file for more details.

"""Circulation exceptions."""

import json

from flask import g
from invenio_rest.errors import RESTException


class CirculationException(RESTException):
    """Base Exception for circulation module, inherit, don't raise."""

    code = 400

    @property
    def name(self):
        """The status name."""
        return type(self).__name__

    def get_body(self, environ=None):
        """Get the request body."""
        body = dict(
            status=self.code,
            message=self.get_description(environ),
            error_module="Circulation",
            error_class=self.name,
        )

        errors = self.get_errors()
        if self.errors:
            body["errors"] = errors

        if self.code and (self.code >= 500) and hasattr(g, "sentry_event_id"):
            body["error_id"] = str(g.sentry_event_id)

        return json.dumps(body)


# Permissions
class InvalidPermissionError(CirculationException):
    """Raised when permissions are not satisfied for transition."""

    code = 403

    def __init__(self, permission=None, **kwargs):
        """Initialize exception."""
        self.description = (
            "This action is not permitted "
            "for your role '{}'".format(permission)
        )
        super().__init__(**kwargs)


# Transitions
class TransitionConstraintsViolationError(CirculationException):
    """Exception raised when constraints for the transition failed."""

    _default = "Transition constraints have failed."


class TransitionConditionsFailedError(CirculationException):
    """Conditions for the transition failed at loan state."""

    description = "Transition conditions have failed."


class NoValidTransitionAvailableError(CirculationException):
    """Exception raised when all transitions conditions failed."""

    def __init__(self, loan_pid=None, state=None, **kwargs):
        """Initialize exception."""
        self.description = (
            "For the loan #'{0}' there are no valid transitions from "
            "its current state '{1}'".format(loan_pid, state)
        )
        super().__init__(**kwargs)


# Loans
class InvalidLoanStateError(CirculationException):
    """State not found in circulation configuration."""

    def __init__(self, state=None, **kwargs):
        """Initialize exception."""
        self.description = "Invalid loan state '{}'".format(state)
        super().__init__(**kwargs)


class ItemNotAvailableError(CirculationException):
    """Exception raised from action on unavailable item."""

    def __init__(self, item_pid=None, transition=None, **kwargs):
        """Initialize exception.

        :param item_pid: a dict containing `value` and `type` fields to
            uniquely identify the item.
        :param transition: the transition that failed.
        """
        self.description = "The item requested with PID '{0}:{1}' is not " \
                           "available. Transition to '{2}' has failed." \
            .format(item_pid["type"], item_pid["value"], transition)
        super().__init__(**kwargs)


class DocumentNotAvailableError(CirculationException):
    """Exception raised from action on unavailable document."""

    def __init__(self, document_pid=None, transition=None, **kwargs):
        """Initialize exception."""
        self.description = (
            "The document requested with PID '{0}' is not available. "
            "Transition to '{1}' has failed.".format(document_pid, transition)
        )
        super().__init__(**kwargs)


class ItemDoNotMatchError(CirculationException):
    """Exception raised from action on unavailable item."""


class DocumentDoNotMatchError(CirculationException):
    """Exception raised from action on unavailable document."""


class MultipleLoansOnItemError(CirculationException):
    """Exception raised when more than one loan on an item."""

    def __init__(self, item_pid=None, **kwargs):
        """Initialize exception.

        :param item_pid: a dict containing `value` and `type` fields to
            uniquely identify the item.
        """
        self.description = "Multiple active loans on item with PID '{0}:{1}'" \
            .format(item_pid["type"], item_pid["value"])
        super().__init__(**kwargs)


class LoanMaxExtensionError(CirculationException):
    """Exception raised when reached the max extensions for a loan."""

    def __init__(self, loan_pid=None, extension_count=None, **kwargs):
        """Initialize exception."""
        self.description = (
            "You have reached the maximum amount of extensions '{}' "
            "for loan '{}'".format(extension_count, loan_pid)
        )
        super().__init__(**kwargs)


# General
class RecordCannotBeRequestedError(CirculationException):
    """Exception raised when item can not be requested."""


class NotImplementedConfigurationError(CirculationException):
    """Exception raised when function is not implemented."""

    description = (
        "Function is not implemented. Implement this function in your module "
        "and pass it to the config variable"
    )

    def __init__(self, config_variable=None, **kwargs):
        """Initialize exception."""
        self.description = "{} '{}'".format(self.description, config_variable)
        super().__init__(**kwargs)


class MissingRequiredParameterError(CirculationException):
    """Exception raised when required parameter is missing."""



# -*- coding: utf-8 -*-
#
# Copyright (C) 2018-2020 CERN.
# Copyright (C) 2018-2020 RERO.
#
# Invenio-Circulation is free software; you can redistribute it and/or modify
# it under the terms of the MIT License; see LICENSE file for more details.

"""Circulation API."""

from elasticsearch import VERSION as ES_VERSION
from flask import current_app
from invenio_jsonschemas import current_jsonschemas
from invenio_pidstore.resolver import Resolver
from invenio_records.api import Record

from .errors import MissingRequiredParameterError, MultipleLoansOnItemError
from .pidstore.pids import CIRCULATION_LOAN_PID_TYPE
from .search.api import search_by_patron_item_or_document, search_by_pid
from .utils import str2datetime


class Loan(Record):
    """Loan record class."""

    DATE_FIELDS = [
        "start_date",
        "end_date",
        "request_expire_date",
        "request_start_date",
    ]
    DATETIME_FIELDS = ["transaction_date"]

    _schema = "loans/loan-v1.0.0.json"

    def __init__(self, data, model=None):
        """Constructor."""
        self.item_ref_builder = current_app.config[
            "CIRCULATION_ITEM_REF_BUILDER"]
        self["state"] = current_app.config["CIRCULATION_LOAN_INITIAL_STATE"]
        super().__init__(data, model)

    @classmethod
    def build_resolver_fields(cls, data):
        """Build all resolver fields."""
        item_ref = current_app.config["CIRCULATION_ITEM_REF_BUILDER"]
        data["item"] = item_ref(data["pid"], data)
        patron_ref = current_app.config["CIRCULATION_PATRON_REF_BUILDER"]
        data["patron"] = patron_ref(data["pid"], data)
        document_ref = current_app.config["CIRCULATION_DOCUMENT_REF_BUILDER"]
        data["document"] = document_ref(data["pid"], data)

    @classmethod
    def create(cls, data, id_=None, **kwargs):
        """Create Loan record."""
        data["$schema"] = current_jsonschemas.path_to_url(cls._schema)
        cls.build_resolver_fields(data)

        # resolve document if `item_pid` provided
        if data.get("item_pid"):
            data["document_pid"] = get_document_pid_by_item_pid(
                data["item_pid"])

        return super().create(data, id_=id_, **kwargs)

    def update(self, *args, **kwargs):
        """Update Loan record."""
        super().update(*args, **kwargs)
        self.build_resolver_fields(self)

    def date_fields2datetime(self):
        """Convert string datetime fields to Python datetime."""
        for field in self.DATE_FIELDS + self.DATETIME_FIELDS:
            if field in self:
                self[field] = str2datetime(self[field])

    def date_fields2str(self):
        """Convert Python datetime fields to string."""
        for field in self.DATE_FIELDS:
            if field in self:
                self[field] = self[field].date().isoformat()
        for field in self.DATETIME_FIELDS:
            if field in self:
                self[field] = self[field].isoformat()

    @classmethod
    def get_record_by_pid(cls, pid, with_deleted=False):
        """Get ils record by pid value."""
        resolver = Resolver(
            pid_type=CIRCULATION_LOAN_PID_TYPE,
            object_type="rec",
            getter=cls.get_record,
        )
        _, record = resolver.resolve(str(pid))
        return record

    def update_item_ref(self, item_pid):
        """Replace item reference.

        :param item_pid: a dict containing `value` and `type` fields to
            uniquely identify the item.
        """
        if not item_pid:
            msg = "Missing required arg 'item_pid' when updating loan '{}'"
            raise MissingRequiredParameterError(
                description=msg.format(self["pid"])
            )
        self["item_pid"] = item_pid


def is_item_available_for_checkout(item_pid):
    """Return True if the given item is available for loan, False otherwise.

    :param item_pid: a dict containing `value` and `type` fields to
        uniquely identify the item.
    """
    config = current_app.config
    cfg_item_can_circulate = config["CIRCULATION_POLICIES"]["checkout"].get(
        "item_can_circulate"
    )
    if not cfg_item_can_circulate(item_pid):
        return False

    search = search_by_pid(
        item_pid=item_pid,
        filter_states=config.get("CIRCULATION_STATES_LOAN_ACTIVE"),
    )
    search_result = search.execute()
    if ES_VERSION[0] >= 7:
        return search_result.hits.total.value == 0
    else:
        return search_result.hits.total == 0


def is_item_at_desk_available_for_checkout(item_pid, patron_pid):
    """Return True if the ITEM_AT_DESK is available for loan, False instead."""
    # Try first to accept the availability for given patron
    item_exception_loans = 0
    search = search_by_patron_item_or_document(
        item_pid=item_pid,
        patron_pid=patron_pid,
        filter_states=["ITEM_AT_DESK"],
    )
    search_result = search.execute()
    if ES_VERSION[0] >= 7:
        item_exception_loans = search_result.hits.total.value
    else:
        item_exception_loans = search_result.hits.total

    if item_exception_loans == 1:
        return True

    # Then use normal behaviour if it failed with patron
    return is_item_available_for_checkout(item_pid)


def can_be_requested(loan):
    """Return True if the given record can be requested, False otherwise."""
    config = current_app.config
    cfg_can_be_requested = config["CIRCULATION_POLICIES"]["request"].get(
        "can_be_requested"
    )
    return cfg_can_be_requested(loan)


def get_pending_loans_by_item_pid(item_pid):
    """Return any pending loans for the given item.

    :param item_pid: a dict containing `value` and `type` fields to
        uniquely identify the item.
    """
    search = search_by_pid(
        item_pid=item_pid,
        filter_states=current_app.config["CIRCULATION_STATES_LOAN_REQUEST"]
    )
    for result in search.scan():
        yield Loan.get_record_by_pid(result["pid"])


def get_pending_loans_by_doc_pid(document_pid):
    """Return any pending loans for the given document."""
    search = search_by_pid(
        document_pid=document_pid,
        filter_states=current_app.config.get(
            "CIRCULATION_STATES_LOAN_REQUEST"
        ),
    )
    for result in search.scan():
        yield Loan.get_record_by_pid(result["pid"])


def get_available_item_by_doc_pid(document_pid):
    """Return an item pid available for this document."""
    for item_pid in get_items_by_doc_pid(document_pid):
        if is_item_available_for_checkout(item_pid):
            return item_pid
    return None


def get_items_by_doc_pid(document_pid):
    """Return a list of item PIDs for this document."""
    return current_app.config["CIRCULATION_ITEMS_RETRIEVER_FROM_DOCUMENT"](
        document_pid
    )


def get_document_pid_by_item_pid(item_pid):
    """Return the document pid of this item_pid."""
    return current_app.config["CIRCULATION_DOCUMENT_RETRIEVER_FROM_ITEM"](
        item_pid
    )


def get_loan_for_item(item_pid):
    """Return the Loan attached to the given item, if any.

    :param item_pid: a dict containing `value` and `type` fields to
        uniquely identify the item.
    """
    if not item_pid:
        return

    search = search_by_pid(
        item_pid=item_pid,
        filter_states=current_app.config["CIRCULATION_STATES_LOAN_ACTIVE"],
    )
    loan = None
    hits = list(search.scan())
    if hits:
        if len(hits) > 1:
            raise MultipleLoansOnItemError(item_pid=item_pid)
        loan = Loan.get_record_by_pid(hits[0]["pid"])
    return loan



# -*- coding: utf-8 -*-
#
# Copyright (C) 2018 CERN.
# Copyright (C) 2018 RERO.
#
# Invenio-Circulation is free software; you can redistribute it and/or modify
# it under the terms of the MIT License; see LICENSE file for more details.

"""Tests for circulation state machine logic."""

import pytest

from invenio_circulation.errors import NoValidTransitionAvailableError
from invenio_circulation.proxies import current_circulation


def test_invalid_transitions(loan_created, app, params):
    """Test that there are no conditional transitions at this state."""
    with pytest.raises(NoValidTransitionAvailableError):
        current_circulation.circulation.trigger(loan_created, **params)



# -*- coding: utf-8 -*-
#
# Copyright (C) 2018-2020 CERN.
# Copyright (C) 2018-2020 RERO.
#
# Invenio-Circulation is free software; you can redistribute it and/or modify
# it under the terms of the MIT License; see LICENSE file for more details.

"""Tests for loan states."""

from invenio_circulation.api import Loan
from invenio_circulation.proxies import current_circulation

from .helpers import SwappedConfig


def test_loan_request_on_document_with_auto_available_item_assignment(
    loan_created, params, mock_is_item_available_for_checkout
):
    """Test loan request on document with auto available item assignment."""
    mock_is_item_available_for_checkout.return_value = True

    # we have a request just on document_pid
    del params["item_pid"]

    # mock functions for automatically assigning an available item
    # and a default pickup location
    other_pid = dict(type="itemid", value="other_pid")
    with SwappedConfig(
        "CIRCULATION_ITEMS_RETRIEVER_FROM_DOCUMENT", lambda x: [other_pid]
    ):
        with SwappedConfig(
            "CIRCULATION_ITEM_LOCATION_RETRIEVER",
            lambda x: "pickup_location_pid",
        ):
            loan = current_circulation.circulation.trigger(
                loan_created, **dict(params, trigger="request")
            )

    assert loan["state"] == "PENDING"
    assert loan["pickup_location_pid"] == "pickup_location_pid"
    assert loan["item_pid"] == other_pid
    assert loan["transaction_date"]


def test_loan_request_on_available_item_with_pickup(loan_created, params):
    """Test loan request action on available item with pickup override."""

    with SwappedConfig(
        "CIRCULATION_ITEM_LOCATION_RETRIEVER", lambda x: "pickup_location_pid"
    ):
        loan = current_circulation.circulation.trigger(
            loan_created,
            **dict(
                params,
                trigger="request",
                pickup_location_pid="other_location_pid",
            )
        )
    assert loan["state"] == "PENDING"
    assert loan["pickup_location_pid"] == "other_location_pid"
    assert loan["item_pid"] == params["item_pid"]


def test_loan_request_on_available_item_default_location(loan_created, params):
    """Test loan request action on available item without pickup override."""

    with SwappedConfig(
        "CIRCULATION_ITEM_LOCATION_RETRIEVER", lambda x: "pickup_location_pid"
    ):
        loan = current_circulation.circulation.trigger(
            loan_created, **dict(params, trigger="request")
        )
    assert loan["state"] == "PENDING"
    assert loan["pickup_location_pid"] == "pickup_location_pid"
    assert loan["item_pid"] == params["item_pid"]


def test_loan_request_on_document_with_unavailable_items(
    loan_created, params, mock_is_item_available_for_checkout
):
    """Test loan request on a document that has no items available."""
    mock_is_item_available_for_checkout.return_value = False

    # we have a request just on document_pid
    del params["item_pid"]

    # find an item attached to the document, which will be unavailable
    with SwappedConfig(
        "CIRCULATION_ITEMS_RETRIEVER_FROM_DOCUMENT", lambda x: ["item_pid"]
    ):
        loan = current_circulation.circulation.trigger(
            loan_created,
            **dict(
                params,
                trigger="request",
                pickup_location_pid="pickup_location_pid",
            )
        )
        assert loan["state"] == "PENDING"
        assert "item_pid" not in loan
        assert loan["document_pid"] == "document_pid"
        assert loan["pickup_location_pid"] == "pickup_location_pid"


def test_auto_assignment_of_returned_item_to_pending_document_requests(
    loan_created,
    params,
    mock_is_item_available_for_checkout,
    mock_get_pending_loans_by_doc_pid,
):
    """Test assignment of newly available items to pending doc requests."""
    mock_is_item_available_for_checkout.return_value = True
    mock_get_pending_loans_by_doc_pid.return_value = ["item_pid"]

    # find which document the returned item is associated to
    with SwappedConfig(
        "CIRCULATION_DOCUMENT_RETRIEVER_FROM_ITEM", lambda x: "document_pid"
    ):
        with SwappedConfig(
            "CIRCULATION_ITEM_LOCATION_RETRIEVER", lambda x: "loc_pid"
        ):
            # start a loan on item with pid 'item_pid'
            new_loan = current_circulation.circulation.trigger(
                loan_created,
                **dict(
                    params,
                    trigger="checkout",
                    item_pid=dict(type="itemid", value="item_pid"),
                    pickup_location_pid="loc_pid",
                )
            )
            assert new_loan["state"] == "ITEM_ON_LOAN"

            # create a new loan request on document_pid without items available
            new_loan_created = Loan.create({"pid": "2"})
            # remove item_pid
            params.pop("item_pid")
            pending_loan = current_circulation.circulation.trigger(
                new_loan_created,
                **dict(
                    params,
                    trigger="request",
                    document_pid="document_pid",
                    pickup_location_pid="loc_pid",
                )
            )
            assert pending_loan["state"] == "PENDING"
            # no item available found. Request is created with no item attached
            assert "item_pid" not in pending_loan
            assert pending_loan["document_pid"] == "document_pid"

            # resolve pending document requests to `document_pid`
            mock_get_pending_loans_by_doc_pid.return_value = [pending_loan]

            returned_loan = current_circulation.circulation.trigger(
                new_loan,
                **dict(
                    params,
                    item_pid=dict(type="itemid", value="item_pid"),
                    pickup_location_pid="loc_pid",
                )
            )
            assert returned_loan["state"] == "ITEM_RETURNED"

            # item `item_pid` has been attached to pending loan request on
            # `document_pid` automatically
            assert pending_loan["state"] == "PENDING"
            assert pending_loan["item_pid"] == dict(
                type="itemid", value="item_pid"
            )
            assert pending_loan["document_pid"] == "document_pid"



# -*- coding: utf-8 -*-
#
# Copyright (C) 2018-2020 CERN.
# Copyright (C) 2018-2020 RERO.
#
# Invenio-Circulation is free software; you can redistribute it and/or modify
# it under the terms of the MIT License; see LICENSE file for more details.

"""Invenio Circulation base transitions."""

import copy
from datetime import datetime

import arrow
from flask import current_app
from invenio_db import db

from ..api import Loan, is_item_available_for_checkout
from ..errors import DocumentDoNotMatchError, DocumentNotAvailableError, \
    InvalidLoanStateError, InvalidPermissionError, ItemNotAvailableError, \
    MissingRequiredParameterError, TransitionConditionsFailedError, \
    TransitionConstraintsViolationError
from ..proxies import current_circulation
from ..signals import loan_state_changed
from ..utils import str2datetime


def ensure_same_patron(f):
    """Validate that the patron PID exists and cannot be changed."""
    def inner(self, loan, **kwargs):
        new_patron_pid = kwargs.get("patron_pid")

        if not current_app.config["CIRCULATION_PATRON_EXISTS"](new_patron_pid):
            msg = "Patron '{0}' not found in the system".format(new_patron_pid)
            raise TransitionConstraintsViolationError(description=msg)

        if loan.get("patron_pid") and new_patron_pid != loan["patron_pid"]:
            msg = (
                "Cannot change patron to '{}' while performing an action "
                "on this loan".format(new_patron_pid)
            )
            raise TransitionConstraintsViolationError(description=msg)

        return f(self, loan, **kwargs)
    return inner


def ensure_same_document(f):
    """Validate that the document PID exists and cannot be changed."""
    def inner(self, loan, **kwargs):
        new_doc_pid = kwargs.get("document_pid")

        if not current_app.config["CIRCULATION_DOCUMENT_EXISTS"](new_doc_pid):
            msg = "Document '{0}' not found in the system".format(new_doc_pid)
            raise DocumentNotAvailableError(description=msg)

        if loan.get("document_pid") and new_doc_pid != loan["document_pid"]:
            msg = (
                "Cannot change document to '{}' while performing an action "
                "on this loan".format(new_doc_pid)
            )
            raise DocumentDoNotMatchError(description=msg)

        return f(self, loan, **kwargs)
    return inner


def ensure_required_params(f):
    """Decorator to ensure that all required parameters has been passed."""
    def inner(self, loan, **kwargs):
        missing = [p for p in self.REQUIRED_PARAMS if p not in kwargs]
        if missing:
            msg = "Required input parameters are missing '[{}]'".format(
                missing
            )
            raise MissingRequiredParameterError(description=msg)
        if all(param not in kwargs for param in self.PARTIAL_REQUIRED_PARAMS):
            msg = "One of the required parameters '[{}]' is missing.".format(
                self.PARTIAL_REQUIRED_PARAMS
            )
            raise MissingRequiredParameterError(description=msg)
        return f(self, loan, **kwargs)
    return inner


def has_permission(f):
    """Decorator to check the transition should be manually triggered."""
    def inner(self, loan, **kwargs):
        if self.permission_factory and not self.permission_factory(loan).can():
            raise InvalidPermissionError(
                permission=self.permission_factory(loan)
            )
        return f(self, loan, **kwargs)
    return inner


def check_trigger(f):
    """Decorator to check the transition should be manually triggered."""
    def inner(self, loan, **kwargs):
        if kwargs.get("trigger", "next") != self.trigger:
            msg = "The transition with trigger '{}' does not exist."
            raise TransitionConditionsFailedError(
                description=msg.format(self.trigger)
            )
        return f(self, loan, **kwargs)
    return inner


class Transition(object):
    """A transition object that is triggered on conditions."""

    REQUIRED_PARAMS = [
        "transaction_user_pid",
        "patron_pid",
        "transaction_location_pid",
    ]

    PARTIAL_REQUIRED_PARAMS = ["item_pid", "document_pid"]

    def __init__(
        self, src, dest, trigger="next", permission_factory=None, **kwargs
    ):
        """Init transition object."""
        self.src = src
        self.dest = dest
        self.trigger = trigger
        default_perm = current_app.config[
            "CIRCULATION_LOAN_TRANSITIONS_DEFAULT_PERMISSION_FACTORY"
        ]
        self.permission_factory = permission_factory or default_perm
        self.validate_transition_states()

    def _check_item_before_availability(self, loan):
        """Common check on item before availability check."""
        if "item_pid" not in loan:
            msg = "Item not set for loan #'{}'".format(loan["pid"])
            raise TransitionConstraintsViolationError(description=msg)

        if not current_app.config["CIRCULATION_ITEM_EXISTS"](loan["item_pid"]):
            raise ItemNotAvailableError(
                item_pid=loan["item_pid"], transition=self.dest
            )

    def ensure_item_is_available_for_checkout(self, loan):
        """Validate that an item is available."""
        self._check_item_before_availability(loan)

        if not is_item_available_for_checkout(loan["item_pid"]):
            raise ItemNotAvailableError(
                item_pid=loan["item_pid"], transition=self.dest
            )

    def validate_transition_states(self):
        """Ensure that source and destination states are valid."""
        states = current_app.config["CIRCULATION_LOAN_TRANSITIONS"].keys()
        if not all([self.src in states, self.dest in states]):
            msg = "Source state '{0}' or destination state '{1}' " \
                  "not in [{2}]".format(self.src, self.dest, states)
            raise InvalidLoanStateError(description=msg)

    def _date_fields2datetime(self, kwargs):
        """Convert any extra kwargs string date to Python datetime."""
        for field in Loan.DATE_FIELDS + Loan.DATETIME_FIELDS:
            if field in kwargs:
                if type(kwargs[field]) is not datetime:
                    kwargs[field] = str2datetime(kwargs[field])

    def before(self, loan, initial_loan, **kwargs):
        """Validate input, evaluate conditions and raise if failed."""
        loan.update(kwargs)
        loan.setdefault("transaction_date", arrow.utcnow())

    @check_trigger
    @has_permission
    @ensure_required_params
    @ensure_same_document
    @ensure_same_patron
    def execute(self, loan, **kwargs):
        """Execute before actions, transition and after actions."""
        self._date_fields2datetime(kwargs)
        loan.date_fields2datetime()

        initial_loan = copy.deepcopy(loan)

        self.before(loan, initial_loan, **kwargs)
        loan["state"] = self.dest
        self.after(loan, initial_loan)

    def after(self, loan, initial_loan):
        """Commit record and index."""
        initial_loan.date_fields2str()
        loan.date_fields2str()

        loan.commit()
        db.session.commit()
        current_circulation.loan_indexer().index(loan)

        loan_state_changed.send(
            self,
            initial_loan=initial_loan,
            loan=loan,
            trigger=self.trigger,
        )



# -*- coding: utf-8 -*-
#
# Copyright (C) 2019 CERN.
# Copyright (C) 2019 RERO.
#
# Invenio-Circulation is free software; you can redistribute it and/or modify
# it under the terms of the MIT License; see LICENSE file for more details.

"""Module tests."""

from __future__ import absolute_import, print_function

import json
from datetime import timedelta

import arrow
import pytest
from flask import url_for


def test_rest_loader_valid_params(app, json_headers):
    """Test rest loader and serializer with valid params."""

    with app.test_client() as client:
        url = url_for("invenio_records_rest.loanid_list")

        now = arrow.utcnow()
        start_date = now + timedelta(days=1)
        end_date = now + timedelta(days=30)
        params = dict(
            document_pid="document_pid",
            item_pid=dict(type="itemid", value="1"),
            patron_pid="patron_pid",
            transaction_date=now.isoformat(),
            transaction_location_pid="loc_pid",
            transaction_user_pid="user_pid",
            start_date=start_date.date().isoformat(),
            end_date=end_date.date().isoformat()
        )
        res = client.post(
            url, data=json.dumps(params), headers=json_headers
        )
        assert res.status_code == 201
        loan = res.get_json()
        assert loan['metadata']['transaction_date'] == now.isoformat()

        url = url_for("invenio_records_rest.loanid_item", pid_value='1')
        res = client.get(url, headers=json_headers)
        assert res.status_code == 200
        loan = res.get_json()
        assert loan['metadata']['transaction_date'] == now.isoformat()


@pytest.mark.parametrize(
    "invalid_params",
    [
        {"transaction_date": arrow.utcnow().format('YYYY-MM-DD')},
        {"transaction_date": arrow.utcnow().to('US/Pacific').isoformat()},
        {"start_date": arrow.utcnow().isoformat()},
        {"end_date": arrow.utcnow().isoformat()},
        {"request_expire_date": arrow.utcnow().isoformat()},
        {"item_pid": dict(value="1")},
        {"item_pid": dict(type="itemid")}
    ]
)
def test_rest_loader_invalid_transaction_date_format(
    app, json_headers, invalid_params
):
    """Test rest loader transaction_date format validation."""

    with app.test_client() as client:
        url = url_for("invenio_records_rest.loanid_list")

        params = dict(
            patron_pid="patron_pid",
            transaction_location_pid="loc_pid",
            transaction_user_pid="user_pid",
            **invalid_params
        )
        res = client.post(
            url, data=json.dumps(params), headers=json_headers
        )
        assert res.status_code == 400
        response = res.get_json()
        assert response['message'] == "Validation error."



# -*- coding: utf-8 -*-
#
# Copyright (C) 2018-2020 CERN.
# Copyright (C) 2018-2020 RERO.
#
# Invenio-Circulation is free software; you can redistribute it and/or modify
# it under the terms of the MIT License; see LICENSE file for more details.

"""Tests for loan states."""

import pytest

from invenio_circulation.errors import NotImplementedConfigurationError, \
    NoValidTransitionAvailableError
from invenio_circulation.proxies import current_circulation
from invenio_circulation.utils import \
    validate_item_pickup_transaction_locations

from .helpers import SwappedConfig


def test_validate_pickup_transaction_locations_not_implemented(
        loan_created, params):
    """Test transition from PENDING to ITEM_AT_DESK.

    when the method validate_item_pickup_transaction_locations is not
    implemented, a NotImplementedConfigurationError is raised.
    """
    loan = current_circulation.circulation.trigger(
        loan_created,
        **dict(
            params,
            trigger="request",
            pickup_location_pid="pickup_location_pid",
        )
    )
    assert loan["state"] == "PENDING"

    with pytest.raises(NotImplementedConfigurationError):
        with SwappedConfig(
            "CIRCULATION_LOAN_LOCATIONS_VERIFICATIONS",
                validate_item_pickup_transaction_locations(
                    loan, "ITEM_AT_DESK")
        ):
            loan = current_circulation.circulation.trigger(
                loan, **dict(params)
            )


def test_validate_item_at_desk(loan_created, params):
    """Test transition from PENDING to ITEM_AT_DESK state."""

    loan = current_circulation.circulation.trigger(
        loan_created,
        **dict(
            params,
            trigger="request",
            pickup_location_pid="pickup_location_pid",
        )
    )
    assert loan["state"] == "PENDING"

    with SwappedConfig(
        "CIRCULATION_ITEM_LOCATION_RETRIEVER", lambda x: "pickup_location_pid"
    ):
        loan = current_circulation.circulation.trigger(
            loan, **dict(params)
        )
        assert loan["state"] == "ITEM_AT_DESK"


def test_validate_item_in_transit_pickup(loan_created, params):
    """Test transition from PENDING to ITEM_IN_TRANSIT_FOR_PICKUP state."""

    loan = current_circulation.circulation.trigger(
        loan_created,
        **dict(
            params,
            trigger="request",
            pickup_location_pid="pickup_location_pid",
        )
    )
    assert loan["state"] == "PENDING"

    with SwappedConfig(
        "CIRCULATION_ITEM_LOCATION_RETRIEVER", lambda x: "pickup_location_pid"
    ):
        loan = current_circulation.circulation.trigger(
            loan, **dict(params,
                         pickup_location_pid="other_location_pid")
        )
        assert loan["state"] == "ITEM_IN_TRANSIT_FOR_PICKUP"


def test_checkout_without_item_attached(loan_created, params):
    """Test checkout on PENDING LOAN without item_pid attached."""

    assert loan_created["state"] == "CREATED"

    with SwappedConfig(
        "CIRCULATION_ITEMS_RETRIEVER_FROM_DOCUMENT", lambda x: []
    ):
        loan = current_circulation.circulation.trigger(
            loan_created, **dict(params, trigger="request")
        )

        assert "item_pid" in loan
        assert "pickup_location_pid" in loan
        assert loan["state"] == "PENDING"

        del loan["item_pid"]
        del params["item_pid"]

        with pytest.raises(NoValidTransitionAvailableError):
            current_circulation.circulation.trigger(
                loan, **dict(params, trigger="next")
            )


def test_checkout_with_different_pickup_location(loan_created, params):
    """Test checkout with pickup location different than item location."""

    assert loan_created["state"] == "CREATED"

    with SwappedConfig(
        "CIRCULATION_ITEM_LOCATION_RETRIEVER", lambda x: "pickup_location_pid"
    ):
        loan = current_circulation.circulation.trigger(
            loan_created, **dict(params, trigger="request",
                                 pickup_location_pid="pickup_location_pid")
        )

        assert loan["item_pid"] == dict(type="itemid", value="item_pid")
        assert loan["pickup_location_pid"] == "pickup_location_pid"
        assert loan["state"] == "PENDING"

        current_circulation.circulation.trigger(
            loan, **dict(params, trigger="next",
                         pickup_location_pid="other_location_pid")
        )

        assert loan["state"] == "ITEM_IN_TRANSIT_FOR_PICKUP"


def test_checkout_with_same_pickup_location(loan_created, params):
    """Test checkout with pickup location same as item location."""

    assert loan_created["state"] == "CREATED"

    with SwappedConfig(
        "CIRCULATION_ITEM_LOCATION_RETRIEVER", lambda x: "pickup_location_pid"
    ):
        loan = current_circulation.circulation.trigger(
            loan_created, **dict(params, trigger="request",
                                 pickup_location_pid="pickup_location_pid")
        )

        assert loan["item_pid"] == dict(type="itemid", value="item_pid")
        assert loan["pickup_location_pid"] == "pickup_location_pid"
        assert loan["state"] == "PENDING"

        current_circulation.circulation.trigger(
            loan, **dict(params, trigger="next",
                         pickup_location_pid="pickup_location_pid")
        )

        assert loan["state"] == "ITEM_AT_DESK"



# -*- coding: utf-8 -*-
#
# Copyright (C) 2018-2020 CERN.
# Copyright (C) 2018-2020 RERO.
#
# Invenio-Circulation is free software; you can redistribute it and/or modify
# it under the terms of the MIT License; see LICENSE file for more details.

"""Version information for Invenio-Circulation.

This file is imported by ``invenio_circulation.__init__``,
and parsed by ``setup.py``.
"""

from __future__ import absolute_import, print_function

__version__ = '1.0.0a27'



# -*- coding: utf-8 -*-
#
# Copyright (C) 2018 CERN.
# Copyright (C) 2018 RERO.
#
# Invenio-Circulation is free software; you can redistribute it and/or modify
# it under the terms of the MIT License; see LICENSE file for more details.

"""Links for record serialization."""

from flask import current_app

from .api import Loan
from .views import build_url_action_for_pid


def loan_links_factory(pid, record=None):
    """Generate links for loan."""
    links = {}
    record = record or Loan.get_record_by_pid(pid.pid_value)
    actions = {}
    transitions_config = current_app.config.get(
        'CIRCULATION_LOAN_TRANSITIONS', {}
    )
    for transition in transitions_config.get(record['state']):
        action = transition.get('trigger', 'next')
        actions[action] = build_url_action_for_pid(pid, action)
    links.setdefault('actions', actions)
    return links



# -*- coding: utf-8 -*-
#
# Copyright (C) 2018-2019 CERN.
# Copyright (C) 2018-2019 RERO.
#
# Invenio-Circulation is free software; you can redistribute it and/or modify
# it under the terms of the MIT License; see LICENSE file for more details.

"""Tests for loan JSON schema."""

from copy import deepcopy

from invenio_circulation.proxies import current_circulation


def test_state_checkout_with_loan_pid(
    loan_created, db, params, mock_is_item_available_for_checkout
):
    """Test that created Loan validates after a checkout action."""
    new_params = deepcopy(params)
    new_params['trigger'] = 'checkout'
    loan = current_circulation.circulation.trigger(loan_created, **new_params)
    loan.validate()


def test_indexed_loans(indexed_loans):
    """Test mappings, index creation and loans indexing."""
    assert indexed_loans



# -*- coding: utf-8 -*-
#
# Copyright (C) 2018-2020 CERN.
# Copyright (C) 2018-2020 RERO.
#
# Invenio-Circulation is free software; you can redistribute it and/or modify
# it under the terms of the MIT License; see LICENSE file for more details.

"""Tests for loan states."""

import copy
from datetime import timedelta

import arrow
import pytest
from flask_security import login_user

from invenio_circulation.api import get_loan_for_item, \
    is_item_at_desk_available_for_checkout, is_item_available_for_checkout
from invenio_circulation.errors import ItemDoNotMatchError, \
    LoanMaxExtensionError, NoValidTransitionAvailableError, \
    RecordCannotBeRequestedError, TransitionConstraintsViolationError
from invenio_circulation.proxies import current_circulation
from invenio_circulation.utils import str2datetime

from .helpers import SwappedConfig, SwappedNestedConfig


def test_override_transaction_date(
    mock_get_pending_loans_by_doc_pid,
    loan_created,
    params,
    mock_is_item_available_for_checkout,
):
    """Test override transaction date."""
    mock_get_pending_loans_by_doc_pid.return_value = []
    mock_is_item_available_for_checkout.return_value = True
    assert loan_created["state"] == "CREATED"

    expected_transaction_date = arrow.utcnow() + timedelta(days=10)
    params["transaction_date"] = expected_transaction_date

    loan = current_circulation.circulation.trigger(
        loan_created, **dict(params, trigger="checkout")
    )
    assert loan["state"] == "ITEM_ON_LOAN"
    assert loan["transaction_date"] == expected_transaction_date.isoformat()


def test_loan_checkout_checkin(
    mock_get_pending_loans_by_doc_pid,
    loan_created,
    db,
    params,
    mock_is_item_available_for_checkout,
):
    """Test loan checkout and checkin actions."""

    mock_get_pending_loans_by_doc_pid.return_value = []
    mock_is_item_available_for_checkout.return_value = True
    assert loan_created["state"] == "CREATED"

    loan = current_circulation.circulation.trigger(
        loan_created, **dict(params, trigger="checkout")
    )
    assert loan["state"] == "ITEM_ON_LOAN"

    # set same transaction location to avoid "in transit"
    same_location = params["transaction_location_pid"]
    with SwappedConfig(
        "CIRCULATION_ITEM_LOCATION_RETRIEVER", lambda x: same_location
    ):
        loan = current_circulation.circulation.trigger(loan, **dict(params))
        assert loan["state"] == "ITEM_RETURNED"
        # Keep changes in database to make item available for checkout later
        db.session.commit()


def test_can_change_item_and_loc_pid_before_checkout(
    mock_get_pending_loans_by_doc_pid,
    loan_created,
    db,
    params,
    mock_is_item_available_for_checkout,
):
    """Test that item pid can be changed before the item is on loan."""

    mock_get_pending_loans_by_doc_pid.return_value = []
    mock_is_item_available_for_checkout.return_value = True
    item_pid = params["item_pid"]
    del params["item_pid"]

    # auto assign both available item pid and the location corresponding to it
    with SwappedConfig(
        "CIRCULATION_ITEMS_RETRIEVER_FROM_DOCUMENT", lambda x: [item_pid]
    ):
        with SwappedConfig(
            "CIRCULATION_ITEM_LOCATION_RETRIEVER",
            lambda x: "pickup_location_pid",
        ):
            loan = current_circulation.circulation.trigger(
                loan_created, **dict(params, trigger="request")
            )
    assert loan["state"] == "PENDING"
    assert loan["pickup_location_pid"] == "pickup_location_pid"
    assert loan["item_pid"] == dict(type="itemid", value="item_pid")

    with SwappedConfig(
        "CIRCULATION_ITEM_LOCATION_RETRIEVER", lambda x: "pickup_location_pid"
    ):
        changed_loan = copy.deepcopy(loan)
        changed_loan = current_circulation.circulation.trigger(
            changed_loan,
            **dict(
                params,
                trigger="next",
                item_pid=dict(type="itemid", value="other_item_pid_1"),
                pickup_location_pid="other_location_pid_1",
            )
        )
        db.session.commit()
        assert changed_loan["state"] == "ITEM_IN_TRANSIT_FOR_PICKUP"
        assert changed_loan["item_pid"] == dict(
            type="itemid", value="other_item_pid_1"
        )
        assert changed_loan["pickup_location_pid"] == "other_location_pid_1"

        changed_loan = copy.deepcopy(loan)
        changed_loan = current_circulation.circulation.trigger(
            changed_loan,
            **dict(
                params,
                trigger="next",
                item_pid=dict(type="itemid", value="other_item_pid_2"),
            )
        )
        db.session.commit()
        assert changed_loan["state"] == "ITEM_AT_DESK"
        assert changed_loan["item_pid"] == dict(
            type="itemid", value="other_item_pid_2"
        )
        assert changed_loan["pickup_location_pid"] == "pickup_location_pid"

        changed_loan = copy.deepcopy(loan)
        changed_loan = current_circulation.circulation.trigger(
            changed_loan,
            **dict(
                params,
                trigger="checkout",
                item_pid=dict(type="itemid", value="other_item_pid_3"),
                pickup_location_pid="other_location_pid_3",
            )
        )
        db.session.commit()
        assert changed_loan["state"] == "ITEM_ON_LOAN"
        assert changed_loan["item_pid"] == dict(
            type="itemid", value="other_item_pid_3"
        )
        assert changed_loan["pickup_location_pid"] == "other_location_pid_3"


def test_cannot_change_item_pid_after_checkout(
    mock_get_pending_loans_by_doc_pid,
    loan_created,
    params,
    mock_is_item_available_for_checkout,
):
    """Test that item pid cannot be changed after the item is on loan."""
    mock_get_pending_loans_by_doc_pid.return_value = []
    mock_is_item_available_for_checkout.return_value = True
    del params["item_pid"]

    # auto assign both available item pid and the location corresponding to it

    with SwappedConfig(
        "CIRCULATION_ITEM_LOCATION_RETRIEVER", lambda x: "pickup_location_pid"
    ):
        loan = current_circulation.circulation.trigger(
            loan_created,
            **dict(
                params,
                trigger="checkout",
                item_pid=dict(type="itemid", value="item_pid"),
            )
        )

    assert loan["state"] == "ITEM_ON_LOAN"
    assert loan["item_pid"] == dict(type="itemid", value="item_pid")
    assert loan["pickup_location_pid"] == "pickup_location_pid"

    with SwappedConfig(
        "CIRCULATION_ITEM_LOCATION_RETRIEVER", lambda x: "pickup_location_pid"
    ):
        changed_loan = copy.deepcopy(loan)
        with pytest.raises(ItemDoNotMatchError):
            current_circulation.circulation.trigger(
                changed_loan,
                **dict(
                    params,
                    trigger="next",
                    item_pid=dict(type="itemid", value="other_item_pid_1"),
                )
            )

        changed_loan = copy.deepcopy(loan)
        with pytest.raises(ItemDoNotMatchError):
            current_circulation.circulation.trigger(
                changed_loan,
                **dict(
                    params,
                    trigger="extend",
                    item_pid=dict(type="itemid", value="other_item_pid_2"),
                )
            )

        changed_loan = copy.deepcopy(loan)
        with pytest.raises(ItemDoNotMatchError):
            current_circulation.circulation.trigger(
                changed_loan,
                **dict(
                    params,
                    trigger="cancel",
                    item_pid=dict(type="itemid", value="other_item_pid_3"),
                )
            )


def test_loan_extend(
    loan_created, params, mock_ensure_item_is_available_for_checkout
):
    """Test loan extend action."""

    mock_ensure_item_is_available_for_checkout.side_effect = None

    loan = current_circulation.circulation.trigger(
        loan_created, **dict(params, trigger="checkout")
    )

    end_date = str2datetime(loan["end_date"])
    default_extension_duration = timedelta(days=30)
    end_date_with_ext = end_date + default_extension_duration

    with SwappedNestedConfig(
        ["CIRCULATION_POLICIES", "extension", "max_count"], lambda x: 0
    ):
        with pytest.raises(LoanMaxExtensionError):
            loan = current_circulation.circulation.trigger(
                loan, **dict(params, trigger="extend")
            )

    with SwappedNestedConfig(
        ["CIRCULATION_POLICIES", "extension", "max_count"], lambda x: 2
    ):
        loan = current_circulation.circulation.trigger(
            loan, **dict(params, trigger="extend")
        )

        new_end_date = loan["end_date"]
        assert str2datetime(new_end_date) == end_date_with_ext
        assert loan["extension_count"] == 1

        loan = current_circulation.circulation.trigger(
            loan, **dict(params, trigger="extend")
        )

        # test too many extensions
        with pytest.raises(LoanMaxExtensionError):
            loan = current_circulation.circulation.trigger(
                loan, **dict(params, trigger="extend")
            )


def test_loan_extend_from_transaction_date(
    loan_created, params, mock_ensure_item_is_available_for_checkout
):
    """Test loan extend action from transaction date."""

    mock_ensure_item_is_available_for_checkout.side_effect = None

    loan = current_circulation.circulation.trigger(
        loan_created, **dict(params, trigger="checkout")
    )
    extension_date = str2datetime(loan["transaction_date"])
    with SwappedNestedConfig(
        ["CIRCULATION_POLICIES", "extension", "from_end_date"], False
    ):
        loan = current_circulation.circulation.trigger(
            loan, **dict(params, trigger="extend")
        )

    default_extension_duration = timedelta(days=30)
    end_date_with_ext = extension_date + default_extension_duration
    assert loan["end_date"] == end_date_with_ext.date().isoformat()
    assert loan["extension_count"] == 1


def test_cancel_action(
    loan_created, params, mock_ensure_item_is_available_for_checkout
):
    """Test should pass when calling `cancel` from `ITEM_ON_LOAN`."""

    mock_ensure_item_is_available_for_checkout.side_effect = None

    loan = current_circulation.circulation.trigger(
        loan_created, **dict(params, trigger="checkout")
    )

    loan = current_circulation.circulation.trigger(
        loan, **dict(params, trigger="cancel")
    )
    assert loan["state"] == "CANCELLED"


def test_cancel_fail(loan_created, params):
    """Test should fail when calling `cancel` from `CREATED`."""
    with pytest.raises(NoValidTransitionAvailableError):
        current_circulation.circulation.trigger(
            loan_created, **dict(params, trigger="cancel")
        )


def test_validate_item_in_transit_for_pickup(loan_created, params):
    """Test transit item in different location."""
    loan = current_circulation.circulation.trigger(
        loan_created,
        **dict(
            params,
            trigger="request",
            pickup_location_pid="pickup_location_pid",
        )
    )
    assert loan["state"] == "PENDING"

    with SwappedConfig(
        "CIRCULATION_ITEM_LOCATION_RETRIEVER",
        lambda x: "external_location_pid",
    ):
        loan = current_circulation.circulation.trigger(loan, **dict(params))
        assert loan["state"] == "ITEM_IN_TRANSIT_FOR_PICKUP"


def test_checkout_extend_with_timedelta(
    loan_created, params, mock_ensure_item_is_available_for_checkout
):
    """Test checkout with a timedelta duration."""
    mock_ensure_item_is_available_for_checkout.side_effect = None

    duration = timedelta(hours=4)
    with SwappedNestedConfig(
        ["CIRCULATION_POLICIES", "checkout", "duration_default"],
        lambda _loan, _initial: duration,
    ):
        loan = current_circulation.circulation.trigger(
            loan_created, **dict(params, trigger="checkout")
        )

        expected_start_date = str2datetime(loan["transaction_date"])
        expected_end_date = str2datetime(loan["start_date"]) + duration
        assert loan["state"] == "ITEM_ON_LOAN"
        assert loan["start_date"] == expected_start_date.date().isoformat()
        assert loan["end_date"] == expected_end_date.date().isoformat()
        end_date = loan["end_date"]

    duration = timedelta(days=1)
    with SwappedNestedConfig(
        ["CIRCULATION_POLICIES", "extension", "duration_default"],
        lambda _loan, _initial: duration,
    ):
        # perform an extension
        loan = current_circulation.circulation.trigger(
            loan, **dict(params, trigger="extend")
        )
        expected_end_date = str2datetime(end_date) + duration
        assert loan["end_date"] == expected_end_date.date().isoformat()
        end_date = loan["end_date"]

    duration = timedelta(days=3, minutes=33)
    with SwappedNestedConfig(
        ["CIRCULATION_POLICIES", "extension", "duration_default"],
        lambda _loan, _initial: duration,
    ):
        # extend again
        loan = current_circulation.circulation.trigger(
            loan, **dict(params, trigger="extend")
        )
        expected_end_date = str2datetime(end_date) + duration
        assert loan["end_date"] == expected_end_date.date().isoformat()


def test_checkout_start_is_transaction_date(
    loan_created, params, mock_ensure_item_is_available_for_checkout
):
    """Test checkout start date to transaction date when not set."""
    mock_ensure_item_is_available_for_checkout.side_effect = None

    duration = timedelta(days=10)
    with SwappedNestedConfig(
        ["CIRCULATION_POLICIES", "checkout", "duration_default"],
        lambda _loan, _initial: duration,
    ):
        loan = current_circulation.circulation.trigger(
            loan_created, **dict(params, trigger="checkout")
        )

        expected_start_date = str2datetime(loan["transaction_date"])
        expected_end_date = str2datetime(loan["start_date"]) + duration
        assert loan["state"] == "ITEM_ON_LOAN"
        assert loan["start_date"] == expected_start_date.date().isoformat()
        assert loan["end_date"] == expected_end_date.date().isoformat()


def test_checkout_with_input_start_end_dates(
    loan_created, params, mock_ensure_item_is_available_for_checkout
):
    """Test checkout start and end dates are set as input."""

    mock_ensure_item_is_available_for_checkout.side_effect = None

    start_date = arrow.utcnow()
    end_date = start_date + timedelta(days=10)
    loan = current_circulation.circulation.trigger(
        loan_created,
        **dict(
            params,
            start_date=start_date,
            end_date=end_date,
            trigger="checkout",
        )
    )
    assert loan["state"] == "ITEM_ON_LOAN"
    assert loan["start_date"] == start_date.date().isoformat()
    assert loan["end_date"] == end_date.date().isoformat()


def test_checkout_fails_when_duration_invalid(
    loan_created, params, mock_ensure_item_is_available_for_checkout
):
    """Test checkout fails when wrong max duration."""

    mock_ensure_item_is_available_for_checkout.side_effect = None

    with pytest.raises(TransitionConstraintsViolationError):
        with SwappedNestedConfig(
            ["CIRCULATION_POLICIES", "checkout", "duration_validate"],
            lambda x: False,
        ):
            start_date = arrow.utcnow()
            end_date = start_date + timedelta(days=60)
            current_circulation.circulation.trigger(
                loan_created,
                **dict(
                    params,
                    start_date=start_date,
                    end_date=end_date,
                    trigger="checkout",
                )
            )


def test_checkin_end_date_is_transaction_date(
    mock_get_pending_loans_by_doc_pid,
    mock_ensure_item_is_available_for_checkout,
    loan_created,
    params,
):
    """Test date the checkin date is the transaction date."""
    mock_get_pending_loans_by_doc_pid.return_value = []
    mock_ensure_item_is_available_for_checkout.side_effect = None

    start_date = arrow.utcnow()
    end_date = start_date + timedelta(days=10)

    loan = current_circulation.circulation.trigger(
        loan_created,
        **dict(
            params,
            start_date=start_date,
            end_date=end_date,
            trigger="checkout",
        )
    )
    assert loan["state"] == "ITEM_ON_LOAN"
    assert loan["transaction_date"]
    transaction_date = loan["transaction_date"]

    same_location = params["transaction_location_pid"]
    params["transaction_date"] = arrow.utcnow()
    with SwappedConfig(
        "CIRCULATION_ITEM_LOCATION_RETRIEVER", lambda x: same_location
    ):
        loan = current_circulation.circulation.trigger(loan, **dict(params))

    assert loan["state"] == "ITEM_RETURNED"
    expected_end_date = str2datetime(transaction_date).date().isoformat()
    assert loan["end_date"] == expected_end_date


def test_item_availability(indexed_loans):
    """Test item_availability with various conditions."""
    assert is_item_available_for_checkout(
        item_pid=dict(type="itemid", value="item_pending_1")
    )
    assert not is_item_available_for_checkout(
        item_pid=dict(type="itemid", value="item_on_loan_2")
    )
    assert is_item_available_for_checkout(
        item_pid=dict(type="itemid", value="item_returned_3")
    )
    assert not is_item_available_for_checkout(
        item_pid=dict(type="itemid", value="item_in_transit_4")
    )
    assert not is_item_available_for_checkout(
        item_pid=dict(type="itemid", value="item_pending_on_loan_6")
    )
    assert is_item_available_for_checkout(
        item_pid=dict(type="itemid", value="item_returned_6")
    )
    assert is_item_available_for_checkout(
        item_pid=dict(type="itemid", value="no_loan")
    )
    # item is not available because it has a loan with state ITEM_AT_DESK
    assert not is_item_available_for_checkout(
        item_pid=dict(type="itemid", value="item_at_desk_5")
    )
    # item is available because the checked-out patron owns the active loan
    assert is_item_at_desk_available_for_checkout(
            item_pid=dict(type="itemid", value="item_at_desk_5"),
            patron_pid="1"
    )


def test_checkout_item_unavailable_steps(loan_created, params, users):
    """Test checkout attempt on unavailable item."""
    user = users["manager"]
    login_user(user)
    with pytest.raises(NoValidTransitionAvailableError):
        # loan created
        current_circulation.circulation.trigger(
            loan_created, **dict(params, trigger="request")
        )
        # loan pending
        current_circulation.circulation.trigger(
            loan_created, **dict(params, trigger="next")
        )
        loan_created["state"] = "ITEM_ON_LOAN"
        current_circulation.circulation.trigger(loan_created, **dict(params))

        # trying to checkout item already on loan
        current_circulation.circulation.trigger(
            loan_created, **dict(params, trigger="checkout")
        )


def test_deny_request(loan_created, params):
    """Test deny request action."""
    with SwappedNestedConfig(
        ["CIRCULATION_POLICIES", "request", "can_be_requested"],
        lambda x: False,
    ):
        with pytest.raises(RecordCannotBeRequestedError):
            current_circulation.circulation.trigger(
                loan_created,
                **dict(
                    params,
                    trigger="request",
                    pickup_location_pid="pickup_location_pid",
                )
            )



# -*- coding: utf-8 -*-
#
# Copyright (C) 2018-2020 CERN.
# Copyright (C) 2018-2020 RERO.
#
# Invenio-Circulation is free software; you can redistribute it and/or modify
# it under the terms of the MIT License; see LICENSE file for more details.

"""Invenio module for the circulation of bibliographic items."""

import os

from setuptools import find_packages, setup

readme = open('README.rst').read()
history = open('CHANGES.rst').read()

tests_require = [
    'check-manifest>=0.35',
    'coverage>=4.5.3',
    'isort>=4.3.11',
    'invenio-app>=1.2.3',
    'invenio-jsonschemas>=1.0.1',
    'mock>=1.3.0',
    'pydocstyle>=3.0.0',
    'pytest-cov>=2.7.1',
    'pytest-pep8>=1.0.6',
    'pytest>=4.6.4,<5.0.0',
    'pytest-invenio>=1.2.1',
]

invenio_search_version = '1.2.3'
invenio_db_version = '1.0.4'

extras_require = {
    'elasticsearch5': [
        'invenio-search[elasticsearch5]>={}'.format(invenio_search_version)
    ],
    'elasticsearch6': [
        'invenio-search[elasticsearch6]>={}'.format(invenio_search_version)
    ],
    'elasticsearch7': [
        'invenio-search[elasticsearch7]>={}'.format(invenio_search_version),
    ],
    'docs': [
        'Sphinx>=1.8.4,<3'
    ],
    'mysql': [
        'invenio-db[mysql,versioning]>={}'.format(invenio_db_version)
    ],
    'postgresql': [
        'invenio-db[postgresql,versioning]>={}'.format(invenio_db_version)
    ],
    'sqlite': [
        'invenio-db[versioning]>={}'.format(invenio_db_version)
    ],
    'tests': tests_require
}

extras_require['all'] = []
for name, reqs in extras_require.items():
    if name in (
        'mysql',
        'postgresql',
        'sqlite',
        'elasticsearch5',
        'elasticsearch6',
        'elasticsearch7',
    ):
        continue
    extras_require['all'].extend(reqs)

setup_requires = ['Babel>=1.3', 'pytest-runner>=2.6.2']

install_requires = [
    'arrow>=0.15.0',
    'invenio-base>=1.2.1',
    'invenio-access>=1.3.1',
    'invenio-logging>=1.2.1',
    'invenio-pidstore>=1.1.0',
    'invenio-records-rest>=1.6.4',
    'invenio-jsonschemas>=1.0.1',
]

packages = find_packages()


# Get the version string. Cannot be done with import!
g = {}
with open(os.path.join('invenio_circulation', 'version.py'), 'rt') as fp:
    exec(fp.read(), g)
    version = g['__version__']

setup(
    name='invenio-circulation',
    version=version,
    description=__doc__,
    long_description=readme + '\n\n' + history,
    keywords='invenio TODO',
    license='MIT',
    author='CERN',
    author_email='info@inveniosoftware.org',
    url='https://github.com/inveniosoftware/invenio-circulation',
    packages=packages,
    zip_safe=False,
    include_package_data=True,
    platforms='any',
    python_requires='>=3',
    entry_points={
        'invenio_base.apps': [
            'invenio_circulation = invenio_circulation:InvenioCirculation'
        ],
        'invenio_base.api_apps': [
            'invenio_circulation = invenio_circulation:InvenioCirculation'
        ],
        'invenio_base.api_blueprints': [
            'invenio_circulation_loan_actions = '
            'invenio_circulation.views:create_loan_actions_blueprint',
            'invenio_circulation_loan_for_item = '
            'invenio_circulation_loan_replace_item = '
            'invenio_circulation.views:create_loan_replace_item_blueprint',

        ],
        'invenio_i18n.translations': ['messages = invenio_circulation'],
        'invenio_pidstore.fetchers': [
            'loanid = invenio_circulation.pidstore.fetchers:loan_pid_fetcher'
        ],
        'invenio_pidstore.minters': [
            'loanid = invenio_circulation.pidstore.minters:loan_pid_minter'
        ],
        'invenio_jsonschemas.schemas': [
            'loans = invenio_circulation.schemas'
        ],
        'invenio_search.mappings': ['loans = invenio_circulation.mappings'],
        "invenio_records.jsonresolver": [
            "item_resolver = invenio_circulation.records.jsonresolver.item",
            "patron_resolver = invenio_circulation.records.jsonresolver.patron",
            "document_resolver = invenio_circulation.records.jsonresolver.document",

        ],
    },
    extras_require=extras_require,
    install_requires=install_requires,
    setup_requires=setup_requires,
    tests_require=tests_require,
    classifiers=[
        'Environment :: Web Environment',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Topic :: Internet :: WWW/HTTP :: Dynamic Content',
        'Topic :: Software Development :: Libraries :: Python Modules',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.5',
        'Development Status :: 1 - Planning',
    ],
)



# -*- coding: utf-8 -*-
#
# Copyright (C) 2019 CERN.
# Copyright (C) 2019 RERO.
#
# Invenio-Circulation is free software; you can redistribute it and/or modify
# it under the terms of the MIT License; see LICENSE file for more details.

"""Circulation Patron JSON Resolver module."""

import jsonresolver
from werkzeug.routing import Rule


@jsonresolver.hookimpl
def jsonresolver_loader(url_map):
    """Resolve the patron reference."""
    from flask import current_app as app
    resolving_path = app.config.get("CIRCULATION_PATRON_RESOLVING_PATH") or "/"
    url_map.add(Rule(
        resolving_path,
        endpoint=app.config.get('CIRCULATION_PATRON_RESOLVER_ENDPOINT'),
        host=app.config.get('JSONSCHEMAS_HOST')))



# -*- coding: utf-8 -*-
#
# Copyright (C) 2018-2020 CERN.
# Copyright (C) 2018-2020 RERO.
#
# Invenio-Circulation is free software; you can redistribute it and/or modify
# it under the terms of the MIT License; see LICENSE file for more details.

"""Circulation views."""

from copy import deepcopy

from flask import Blueprint, current_app, request, url_for
from invenio_db import db
from invenio_records_rest.utils import obj_or_import_string
from invenio_records_rest.views import pass_record
from invenio_rest import ContentNegotiatedMethodView

from .errors import InvalidLoanStateError, ItemNotAvailableError, \
    MissingRequiredParameterError
from .permissions import need_permissions
from .pidstore.pids import _LOANID_CONVERTER, CIRCULATION_LOAN_PID_TYPE
from .proxies import current_circulation
from .records.loaders import loan_loader, loan_replace_item_loader
from .signals import loan_replace_item


def extract_transitions_from_app(app):
    """Return all possible actions for configured transitions."""
    transitions_config = app.config.get("CIRCULATION_LOAN_TRANSITIONS", {})
    distinct_actions = set()
    for src_state, transitions in transitions_config.items():
        for t in transitions:
            distinct_actions.add(t.get("trigger", "next"))
    return distinct_actions


def build_url_action_for_pid(pid, action):
    """Build urls for Loan actions."""
    return url_for(
        "invenio_circulation_loan_actions.{0}_actions".format(pid.pid_type),
        pid_value=pid.pid_value,
        action=action,
        _external=True,
    )


def _get_loan_endpoint_options(app):
    """Return the configured endpoint options."""
    endpoints = app.config.get("CIRCULATION_REST_ENDPOINTS", [])
    options = deepcopy(endpoints.get(CIRCULATION_LOAN_PID_TYPE, {}))
    default_media_type = options.get("default_media_type", "")
    rec_serializers = options.get("record_serializers", {})
    serializers = {
        mime: obj_or_import_string(func)
        for mime, func in rec_serializers.items()
    }
    return (
        options,
        dict(
            serializers=serializers,
            default_media_type=default_media_type,
            ctx={},
        ),
    )


def create_loan_actions_blueprint(app):
    """Create a blueprint for Loan actions."""
    blueprint = Blueprint(
        "invenio_circulation_loan_actions", __name__, url_prefix=""
    )

    all_options, view_options = _get_loan_endpoint_options(app)
    view_options["ctx"]["loader"] = loan_loader
    loan_actions = LoanActionResource.as_view(
        LoanActionResource.view_name.format(CIRCULATION_LOAN_PID_TYPE),
        **view_options
    )

    distinct_actions = extract_transitions_from_app(app)
    url = "{0}/<any({1}):action>".format(
        all_options["item_route"], ",".join(distinct_actions)
    )

    blueprint.add_url_rule(url, view_func=loan_actions, methods=["POST"])
    return blueprint


class LoanActionResource(ContentNegotiatedMethodView):
    """Loan action resource."""

    view_name = "{0}_actions"

    def __init__(self, serializers, ctx, *args, **kwargs):
        """Constructor."""
        super().__init__(serializers, *args, **kwargs)
        for key, value in ctx.items():
            setattr(self, key, value)

    @need_permissions("loan-actions")
    @pass_record
    def post(self, pid, record, action, **kwargs):
        """Handle loan action."""
        data = self.loader()
        record = current_circulation.circulation.trigger(
            record, **dict(data, trigger=action)
        )
        db.session.commit()
        return self.make_response(
            pid,
            record,
            202,
            links_factory=current_app.config.get(
                "CIRCULATION_LOAN_LINKS_FACTORY"
            ),
        )


def create_loan_replace_item_blueprint(app):
    """Create a blueprint for replacing Loan Item."""
    blueprint = Blueprint(
        "invenio_circulation_loan_replace_item", __name__, url_prefix=""
    )

    _, view_options = _get_loan_endpoint_options(app)
    view_options["ctx"]["loader"] = loan_replace_item_loader
    replace_item_view = LoanReplaceItemResource.as_view(
        LoanReplaceItemResource.view_name.format(CIRCULATION_LOAN_PID_TYPE),
        **view_options
    )

    url = "circulation/loans/<{0}:pid_value>/replace-item".format(
        _LOANID_CONVERTER
    )
    blueprint.add_url_rule(url, view_func=replace_item_view, methods=["POST"])
    return blueprint


def validate_replace_item(loan, new_item_pid):
    """Validate the new item before replacing the item of the loan.

    :param loan: the current loan to modify.
    :param new_item_pid: a dict containing `value` and `type` fields to
        uniquely identify the item.
    """
    active_states = current_app.config["CIRCULATION_STATES_LOAN_ACTIVE"]
    if loan["state"] not in active_states:
        raise InvalidLoanStateError(
            description=(
                "Cannot replace item in a loan that is not in active state. "
                "Current loan state '{}'".format(loan["state"])
            )
        )

    if not new_item_pid:
        raise MissingRequiredParameterError(
            description="Parameter 'new_item_pid' is required."
        )

    item_exists_func = current_app.config["CIRCULATION_ITEM_EXISTS"]
    if not item_exists_func(new_item_pid):
        raise ItemNotAvailableError(item_pid=new_item_pid)


class LoanReplaceItemResource(ContentNegotiatedMethodView):
    """Loan update resource."""

    view_name = "loan_replace_item_resource"

    def __init__(self, serializers, ctx, *args, **kwargs):
        """Constructor."""
        super().__init__(serializers, *args, **kwargs)
        for key, value in ctx.items():
            setattr(self, key, value)

    @need_permissions("loan-actions")
    @pass_record
    def post(self, pid, record, *args, **kwargs):
        """Handle POST request to update loan with new item."""
        data = self.loader()
        old_item_pid = record.get("item_pid")
        new_item_pid = data.get("item_pid")

        validate_replace_item(record, new_item_pid)
        record.update_item_ref(new_item_pid)

        record.commit()
        db.session.commit()
        current_circulation.loan_indexer().index(record)

        if old_item_pid:
            loan_replace_item.send(self, old_item_pid=old_item_pid,
                                   new_item_pid=new_item_pid)

        return self.make_response(
            pid,
            record,
            202,
            links_factory=current_app.config.get(
                "CIRCULATION_LOAN_LINKS_FACTORY"
            ),
        )



# -*- coding: utf-8 -*-
#
# Copyright (C) 2018-2020 CERN.
# Copyright (C) 2018-2020 RERO.
#
# Invenio-Circulation is free software; you can redistribute it and/or modify
# it under the terms of the MIT License; see LICENSE file for more details.

"""Invenio Circulation transitions conditions."""

from flask import current_app


def is_same_location(item_pid, input_location_pid):
    """Return True if item belonging location is same as input parameter.

    :param item_pid: a dict containing `value` and `type` fields to
        uniquely identify the item.
    """
    item_location_pid = current_app.config[
        "CIRCULATION_ITEM_LOCATION_RETRIEVER"
    ](item_pid)
    return input_location_pid == item_location_pid



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2018-2019 CERN.
# Copyright (C) 2018-2019 RERO.
#
# Invenio is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Helper methods for tests."""

import copy
import uuid

from flask import current_app
from invenio_db import db
from invenio_pidstore import current_pidstore
from invenio_records_rest.utils import deny_all
from six.moves import reduce

from invenio_circulation.api import Loan
from invenio_circulation.permissions import has_read_loan_permission
from invenio_circulation.pidstore.pids import CIRCULATION_LOAN_MINTER


class SwappedConfig:
    """Helper to hot-swap a configuration value."""

    def __init__(self, key, value):
        """Constructor."""
        self.key = key
        self.new_value = value

    def __enter__(self):
        """Save previous value and swap it with the new."""
        self.prev_value = current_app.config[self.key]
        current_app.config[self.key] = self.new_value

    def __exit__(self, type, value, traceback):
        """Restore previous value."""
        current_app.config[self.key] = self.prev_value


class SwappedNestedConfig:
    """Helper to hot-swap a nested configuration value."""

    def __init__(self, nested_keys, value):
        """Constructor."""
        self.nested_keys = nested_keys
        self.new_value = value

    def __enter__(self):
        """Save previous value and swap it with the new."""
        config_obj = reduce(dict.__getitem__, self.nested_keys[:-1],
                            current_app.config)
        self.prev_value = config_obj[self.nested_keys[-1]]
        config_obj[self.nested_keys[-1]] = self.new_value

    def __exit__(self, type, value, traceback):
        """Restore previous value."""
        config_obj = reduce(dict.__getitem__, self.nested_keys[:-1],
                            current_app.config)
        config_obj[self.nested_keys[-1]] = self.prev_value


def create_loan(data):
    """Create a test record."""
    with db.session.begin_nested():
        data = copy.deepcopy(data)
        rec_uuid = uuid.uuid4()
        pid = current_pidstore.minters[CIRCULATION_LOAN_MINTER](rec_uuid, data)
        record = Loan.create(data, id_=rec_uuid)
        return pid, record


def test_views_permissions_factory(action):
    """Test views permissions factory."""
    if action == 'loan-read-access':
        return has_read_loan_permission()
    else:
        return deny_all()
